# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from flask import render_template
from monitor_server import __version__ as monitor_server_version
from monitor_server.api.components import (ListComponents, CountComponents, ListMetricsByComponent,
                                           ListComponentPipelines, ListComponentPipelineBuilds,
                                           ListSessionsOfComponentPipelineBuild,
                                           ListComponentMetricsByItem, ListComponentMetricsByVariant)
from monitor_server.api.execution_context import CountContexts, ListContexts
from monitor_server.api.metrics import (ListMetrics, CountMetrics,
                                        ListMetricsByItem, ListMetricsByItemPattern, CountMetricsByItem,
                                        ListMetricsByItemVariant, ListMetricsByItemVariantPattern,
                                        CountMetricsByItemVariant, ListSessionsByScm,
                                        ListMetricsByScm, CountMetricsBySCMRef,
                                        ListMetricsByType, CountMetricsByType)
from monitor_server.api.pipelines import ListPipelines, ListPipelineBuilds, ListSessionsOfBuildPipeline
from monitor_server.api.sessions import CountMetricsBySession, ListSessions, CountSessions, ListMetricsBySession
from monitor_server.api.resources import (ListHeadResource, ListTailResource,
                                          ListHeadResourceOfComponent, ListTailResourceOfComponent,
                                          ListHeadResourceOfPipeline, ListTailResourceOfPipeline,
                                          ListHeadResourceOfBuild, ListTailResourceOfBuild)
from monitor_server.data.model import MetricModel, SessionModel, ExecutionContextModel
from sqlalchemy import func


def _set_generic_entry_points(server):

    @server.route('/index.html')
    @server.route('/')
    def serve_index():
        session = server.DB.session
        metric_count = session.query(func.count(MetricModel.test_id)).first()[0]
        session_count = session.query(func.count(SessionModel.h)).first()[0]
        context_count = session.query(func.count(ExecutionContextModel.h)).first()[0]
        return render_template('index.html', metric_count=metric_count,
                               session_count=session_count, context_count=context_count,
                               version=monitor_server_version)


def _set_resource_access_entry_points(api):
    api.add_resource(CountComponents)
    api.add_resource(CountContexts)
    api.add_resource(CountMetricsByItem)
    api.add_resource(CountMetricsByItemVariant)
    api.add_resource(CountMetricsBySession)
    api.add_resource(CountMetrics)
    api.add_resource(CountMetricsBySCMRef)
    api.add_resource(CountMetricsByType)
    api.add_resource(CountSessions)
    api.add_resource(ListComponents)
    api.add_resource(ListComponentMetricsByItem)
    api.add_resource(ListComponentMetricsByVariant)
    api.add_resource(ListComponentPipelines)
    api.add_resource(ListComponentPipelineBuilds)
    api.add_resource(ListContexts)
    api.add_resource(ListSessionsByScm)
    api.add_resource(ListTailResource)
    api.add_resource(ListTailResourceOfBuild)
    api.add_resource(ListTailResourceOfComponent)
    api.add_resource(ListTailResourceOfPipeline)
    api.add_resource(ListMetrics)
    api.add_resource(ListMetricsByComponent)
    api.add_resource(ListMetricsByItem)
    api.add_resource(ListMetricsByItemPattern)
    api.add_resource(ListMetricsByItemVariant)
    api.add_resource(ListMetricsByItemVariantPattern)
    api.add_resource(ListMetricsByScm)
    api.add_resource(ListMetricsBySession)
    api.add_resource(ListMetricsByType)
    api.add_resource(ListPipelines)
    api.add_resource(ListPipelineBuilds)
    api.add_resource(ListHeadResource)
    api.add_resource(ListHeadResourceOfBuild)
    api.add_resource(ListHeadResourceOfComponent)
    api.add_resource(ListHeadResourceOfPipeline)
    api.add_resource(ListSessionsOfBuildPipeline)
    api.add_resource(ListSessionsOfComponentPipelineBuild)
    api.add_resource(ListSessions)


def set_entry_points(server):
    _set_generic_entry_points(server)
    _set_resource_access_entry_points(server.API_V1)



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from flask import Flask, Blueprint
from flask_sqlalchemy import SQLAlchemy
from flask_restx import Api

import os
import pathlib

__version__ = '1.0.2'


def _create_monitor_server():
    if os.environ.get('MONITOR_SERVER_KEEP_PROXY', None) is None:
        for env in ['http_proxy', 'https_proxy']:  # pragma: no cover
            if env in os.environ:
                del os.environ[env]

    templates = pathlib.Path(__file__).parent / 'templates'
    server = Flask('monitor_server', template_folder=str(templates))

    mode = os.environ.get('FLASK_ENV', 'dev').lower()
    mode = 'dev' if mode not in ['prod', 'test'] else mode
    if mode == "prod" and "MONITOR_SERVER_DATABASE_URI" not in os.environ:  # pragma: no cover
        raise EnvironmentError('Missing path to database. Please configure it through '
                               'MONITOR_SERVER_DATABASE_URI environment variable')
    server.config.from_object(f'monitor_server.config.{mode.capitalize()}Config')
    v1 = Blueprint('v1', __name__, url_prefix='/api/v1')
    api = Api(v1, doc='/docs', validate=True, title='Monitor Server REST Api',
              description='REST Api to help you analyze data from pytest-monitor runs.',
              license='MIT LICENSE', license_url='https://opensource.org/licenses/MIT',
              contact='Jean-Sébastien Dieu', contact_email='jean-sebastien.dieu@cfm.fr')
    server.API_V1 = api
    server.register_blueprint(v1)
    db = SQLAlchemy(server)
    server.DB = db
    return server


SERVER = _create_monitor_server()



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from flask import current_app, url_for
from http import HTTPStatus


PARAM_PAGE = 'page'
PARAM_ITEM_PER_PAGE = 'per_page'
DEFAULT_PAGINATION_COUNT = 500


def get_value_from_request(req, key, default, vtype=int):
    v = req.args.get(key, None, type=vtype)
    if v is None:
        if req.json is not None and key in req.json:
            return req.json[key]
        return default
    return v


def get_page_info(req):
    item_per_page = current_app.config.get('PAGINATION_COUNT', DEFAULT_PAGINATION_COUNT)
    page = get_value_from_request(req, PARAM_PAGE, 1)
    return page, item_per_page


def add_page_info(pagination, url, data, **url_params):
    if pagination.has_prev:
        params = dict(url_params['url_params']) if url_params.get('url_params', None) else dict()
        params.update({f'{PARAM_PAGE}': pagination.prev_num})
        prev_url = url_for(url, **params)
        data.update(dict(prev_url=prev_url))
    if pagination.has_next:
        params = dict(url_params['url_params']) if url_params.get('url_params', None) else dict()
        params.update({f'{PARAM_PAGE}': pagination.next_num})
        next_url = url_for(url, **params)
        data.update(dict(next_url=next_url))
    data.update(dict(total_page=pagination.pages))
    return data


def make_page(request, query, item_envelope=None, collector=None):
    page, count = get_page_info(request)
    if page < 0:
        pager = query.paginate(page=1, per_page=count)
        pager = query.paginate(page=pager.pages + page + 1, per_page=count)
    else:
        pager = query.paginate(page=page, per_page=count)
    if collector is None:
        content = [i for i in pager.items]
    else:
        content = collector(pager.items)
    if item_envelope is not None:
        data = {item_envelope: content}
    else:
        data = content
    args = dict()
    if request.view_args:
        args.update(request.view_args)
    if request.args:
        args.update(request.args)
    d = add_page_info(pager, request.endpoint, data, url_params=args)
    rc = HTTPStatus.OK if len(content) else HTTPStatus.NO_CONTENT
    return d, rc



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from monitor_api.enums import Field
from monitor_api.core.entities import Context, Metric, Session
import collections
import pandas as pd
import typing
import yaml


class Contexts(collections.UserDict):
    """Dict like structure for manipulating `monitor_api.core.entities.Context` objects."""
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def add(self, context: Context) -> Context:
        """Add a context object to this collection.

        :param context: Context object to add. As for dictionaries, overwrites any existing entry with equivalent hash.
        :type context: monitor_api.core.entities.Context
        :return: The object which have been added
        :rtype: monitor_api.core.entities.Context
        """
        self[context.h] = context
        return context

    def filter_with(self, cond: typing.Callable[[Context], bool], inplace: bool = False):
        """Apply a filter to the bag of `monitor_api.core.entities.Context` held by the instance
        calling this method. Only `monitor_api.core.entities.Context` objects matching the condition method
        are kept.

        :param cond: A callable returning a boolean (True means keep) with a unique parameter of type
                     `monitor_api.core.entities.Context`
        :type cond: Callable[[`monitor_api.core.entities.Context`], bool]
        :param inplace: If True, change the content of this instance by keeping only values for which cond is True
        :type inplace: bool
        :return: The object itself
        :rtype: `monitor_api.core.collections.Contexts`
        """
        c = Contexts()
        for context in self.data.values():
            if cond(context):
                c[context.h] = context
        if inplace:
            self.data = c.data
            return self
        return c

    def to_df(self, keep: typing.List[Field] = None, drop: typing.List[Field] = None) -> pd.DataFrame:
        """Turn a collection of `monitor_api.core.entities.Context` into a pandas DataFrame.

         You can keep or drop a given list of `monitor_api.Field` by specifying either *keep* or
         *drop* as a parameter. If both are provided, *keep* will be used instead of *drop*.

        :param keep: List of field to keep during the export. Cannot be used conjointly with *drop*
        :type keep: list(`monitor_api.Field`)
        :param drop: List of field to keep during the export. Cannot be used conjointly with *keep*
        :type drop: list(`monitor_api.Field`)
        :return: A valid pandas DataFrame object representing the collection of this instance, without any
                 index set minus eventual information that might have been marked as dropped.
        :rtype: pandas.DataFrame
        """
        d = []
        for c in self.data.values():
            d.append(c.to_dict(keep=keep, drop=drop))
        return pd.DataFrame(d)

    def __str__(self):
        d = dict(contexts=self.data)
        for i, c in enumerate(self.data.values()):
            cd = c.to_dict()
            d['contexts'][cd['context_h']] = cd
            del d['contexts'][cd['context_h']]['context_h']
        return yaml.dump(d)


class Sessions(collections.UserDict):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def add(self, session: Session) -> Session:
        """Add a session object to this collection.

        :param session: Session object to add. As for dictionaries, overwrites any existing entry with equivalent hash.
        :type session: monitor_api.core.entities.Session
        :return: The object which have been added
        :rtype: monitor_api.core.entities.Session
        """
        self[session.h] = session
        return session

    def filter_with(self, cond: typing.Callable[[Session], bool], inplace: bool = False):
        """Apply a filter to the bag of `monitor_api.core.entities.Session` held by the instance
        calling this method. Only `monitor_api.core.entities.Session` objects matching the condition method
        are kept.

        :param cond: A callable returning a boolean (True means keep) with a unique parameter of type
                     `monitor_api.core.entities.Session`
        :type cond: Callable[[`monitor_api.core.entities.Session`], bool]
        :param inplace: If True, change the content of this instance by keeping only values for which cond is True
        :type inplace: bool
        :return: The object itself
        :rtype: `monitor_api.core.collections.Sessions`
        """
        s = Sessions()
        for session in self.data.values():
            if cond(session):
                s[session.h] = session
        if inplace:
            self.data = s.data
            return self
        return s

    def with_scm(self, scm: str):
        """Performs a filter operation in order to get a new collection with only session objects with the requested
           source code reference.

           :param scm: The source code reference to use for filter operation
           :type scm: str
           :return: A new Sessions object holding only sessions with scm reference set to the requested one.
           :rtype: monitor_api.core.collections.Sessions
        """
        s = Sessions()
        for session in self.data.values():
            if session.scm == scm:
                s[session.h] = session
        return s

    def with_tags(self, *args, **kwargs):
        """Performs a filter operation in order to get a new collection with only session objects with the requested
           tag specification. You can query for tag presence and/or tag value

           :param args: The list of tags to query for presence
           :type args: list[str]
           :param kwargs: Dictionary of tags to query for their values.
           :return: A new Sessions object holding only sessions matching the specified tag requirements.
           :rtype: monitor_api.core.collections.Sessions
        """
        coll = Sessions()
        for session in self.data.values():
            insert_ok = True
            for tag in args:
                if tag not in session.tags:
                    insert_ok = False
                    break
            for tag, value in kwargs.items():
                if tag not in session.tags:
                    insert_ok = False
                    break
                else:
                    if session.tags[tag] != value:
                        insert_ok = False
                        break
            if insert_ok and (args or kwargs):
                coll[session.h] = session
        return coll

    def to_df(self, keep: typing.List[Field] = None, drop: typing.List[Field] = None) -> pd.DataFrame:
        """Turn a collection of `monitor_api.core.entities.Session` into a pandas DataFrame.

         You can keep or drop a given list of `monitor_api.Field` by specifying either *keep* or
         *drop* as a parameter. If both are provided, *keep* will be used instead of *drop*.
         Tags are exported (if requested) using the expand strategy (see `monitor_api.core.entities.Session.to_dict`
         for more details)

        :param keep: List of field to keep during the export. Cannot be used conjointly with *drop*
        :type keep: list(`monitor_api.Field`)
        :param drop: List of field to keep during the export. Cannot be used conjointly with *keep*
        :type drop: list(`monitor_api.Field`)
        :return: A valid pandas DataFrame object representing the collection of this instance, without any
                 index set minus eventual information that might have been marked as dropped.
        :rtype: pandas.DataFrame
        """
        d = []
        for s in self.data.values():
            d.append(s.to_dict(keep=keep, drop=drop, expand_tags=True))
        return pd.DataFrame(d)

    def __str__(self):
        d = dict(sessions=dict())
        for s in self.data.values():
            sd = s.to_dict()
            d['sessions'][sd['session_h']] = sd
            if not d['sessions'][sd['session_h']]['tags']:
                del d['sessions'][sd['session_h']]['tags']
            del d['sessions'][sd['session_h']]['session_h']
        return yaml.dump(d)


class Metrics(collections.UserList):
    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def filter_with(self, cond: typing.Callable[[Metric], bool], inplace: bool = False):
        """Apply a filter to the bag of `monitor_api.core.entities.Metric` held by the instance
        calling this method. Only `monitor_api.core.entities.Metric` objects matching the condition method
        are kept.

        :param cond: A callable returning a boolean (True means keep) with a unique parameter of type
                     `monitor_api.core.entities.Metric`
        :type cond: Callable[[`monitor_api.core.entities.Metric`], bool]
        :param inplace: If True, change the content of this instance by keeping only values for which cond is True
        :type inplace: bool
        :return: The object itself
        :rtype: `monitor_api.core.collections.Metrics`
        """
        m = Metrics()
        for metric in self.data:
            if cond(metric):
                m.append(metric)
        if inplace:
            self.data = m.data
            return self
        return m

    def variants_of(self, item: str):
        """Performs a filter operation in order to get a new collection with only metric objects regarding any
           variant of the provided item.

           :param item: The test item for which you want to collect all variants known by this collection.
           :type item: str
           :return: A new Metrics object holding only metrics concerning a variant of the provided item.
           :rtype: monitor_api.core.collections.Metrics
        """
        m = Metrics()
        for metric in self.data:
            if metric.item == item:
                m.append(metric)
        return m

    def to_df(self, sessions: Sessions = None, contexts: Contexts = None,
              keep: typing.List[Field] = None, drop: typing.List[Field] = None) -> pd.DataFrame:
        """Turn a collection of `monitor_api.core.entities.Session` into a pandas DataFrame.

         You can keep or drop a given list of `monitor_api.Field` by specifying either *keep* or
         *drop* as a parameter. If both are provided, *keep* will be used instead of *drop*.

        :param sessions:
        :type sessions:
        :param contexts:
        :type contexts:
        :param keep: List of field to keep during the export. Cannot be used conjointly with *drop*
        :type keep: list(`monitor_api.Field`)
        :param drop: List of field to keep during the export. Cannot be used conjointly with *keep*
        :type drop: list(`monitor_api.Field`)
        :return: A valid pandas DataFrame object representing the collection of this instance, without any
                 index set minus eventual information that might have been marked as dropped.
        :rtype: pandas.DataFrame
        """

        d = []
        for m in self.data:
            v = m.to_dict(keep=keep, drop=drop)
            if sessions:
                v.update(sessions[m.session].to_dict(keep=keep, drop=drop, expand_tags=True))
            if contexts:
                v.update(contexts[m.context].to_dict(keep=keep, drop=drop))
            d.append(v)
        df = pd.DataFrame(d)
        return df

    @staticmethod
    def merge(metric_collection_r, metric_collection_l):
        """Merge two collection of metrics and ensure that no duplicate remains in the resulting collection.

        :param metric_collection_r: First collection to merge
        :param metric_collection_l: Second collection to merge
        :type metric_collection_l: monitor_api.core.collections.Metrics
        :type metric_collection_r: monitor_api.core.collections.Metrics
        :return: A new collection without duplicates.
        :rtype: monitor_api.core.collections.Metrics
        """
        processed = set()
        m = Metrics()
        for collection in (metric_collection_l, metric_collection_r):
            for metric in collection:
                key = metric.hash()
                if key not in processed:
                    m.append(metric)
                    processed.add(key)
        return m

    def unique(self, inplace: bool = False):
        """Filter the collection so that no duplicated metric remains.

        :param inplace: Modify this collection so that duplicates are removed.
        :type inplace: bool
        :return: A new collection without duplicates.
        :rtype: monitor_api.core.collections.Metrics
        """
        processed = set()
        m = Metrics()
        for metric in self.data:
            key = metric.hash()
            if key not in processed:
                m.append(metric)
                processed.add(key)
        if inplace:
            self.data = m.data
            return self
        return m

    def __str__(self):
        d = dict(metrics=dict())
        for i, m in enumerate(self.data):
            md = m.to_dict()
            d['metrics'][f'metric_{i}'] = md
        return yaml.dump(d)



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from monitor_api.enums import Field, MetricScope, CONTEXT_ALL_FIELDS, METRIC_ALL_FIELDS, SESSION_ALL_FIELDS
from typing import List, Union, Dict, Optional
import datetime
import hashlib


try:
    datetime.datetime.fromisoformat(datetime.datetime.now().isoformat())
    api_fromisoformat = datetime.datetime.fromisoformat
except AttributeError:
    from monitor_api.py36 import fromisoformat
    api_fromisoformat = fromisoformat


class Metric:
    """Basic metric record.

    :param context_h: Context identifier.
    :type context_h: str
    :param session_h: Session identifier.
    :type session_h: str
    :param start_time: Time at which the metrics have been registered (in ISO format).
    :type context_h: str
    :param item_path: Python like path to the item (with eventual modules).
    :type item_path: str
    :param item: Test item name.
    :type item: str
    :param variant: Full item qualified identifier.
    :type variant: str
    :param path: Path on the filesystem to the module containing the test item.
    :type path: str
    :param kind: Metric's scope (function, module).
    :type kind: str
    :param component: Logical system to which the test belongs.
    :type component: str
    :param wall_time: Total time spent running the test.
    :type wall_time: float
    :param user_time: Time spent running the test in user mode.
    :type user_time: float
    :param kernel_time: Time spent running the test in kernel mode.
    :type kernel_time: float
    :param cpu_usage: Cpu usage while running the test.
    :type cpu_usage: float
    :param memory_usage: Total memory used by the test.
    :type memory_usage: float
    """
    def __init__(self, context_h: str = None, session_h: str = None, start_time: str = None,
                 item_path: str = None, item: str = None, variant: str = None, path: str = None,
                 kind: MetricScope = None, component: str = None, wall_time: float = None, user_time: float = None,
                 kernel_time: float = None, cpu_usage: float = None, memory_usage: float = None):
        self.__c_h = context_h or ''
        self.__s_h = session_h or ''
        self.__time = api_fromisoformat(start_time)
        self.__item_path = item_path or ''
        self.__item = item or ''
        self.__variant = variant or ''
        self.__path = path or ''
        self.__kind = MetricScope.FUNCTION if kind is None else kind
        self.__component = component or ''
        self.__wall = wall_time or 0.0
        self.__user = user_time or 0.0
        self.__kernel = kernel_time or 0.0
        self.__cpu = cpu_usage or 0.0
        self.__memory = memory_usage or 0.0

    def is_function(self):
        """Test if the metric scope is positioned to FUNCTION.

        :return: True if the metric concerns a test function
        :rtype: bool
        """
        return self.__kind == MetricScope.FUNCTION

    def is_package(self):
        """Test if the metric scope is positioned to PACKAGE.

        :return: True if the metric concerns a test package
        :rtype: bool
        """
        return self.__kind == MetricScope.PACKAGE

    def is_module(self):
        """Test if the metric scope is positioned to MODULE.

        :return: True if the metric concerns a test module
        :rtype: bool
        """
        return self.__kind == MetricScope.MODULE

    @property
    def context(self) -> str:
        """Retrieves the context identifier linked to this metric.

        :return: A context identifier
        :rtype: str
        """
        return self.__c_h

    @property
    def session(self) -> str:
        """Retrieves the session identifier linked to this metric.

        :return: A session identifier
        :rtype: str
        """
        return self.__s_h

    @property
    def run_time(self) -> datetime.datetime:
        """Provide the time at which the metrics have been measured.

        :return: A date time expressed in UTC Coordinates.
        :rtype: datetime.datetime
        """
        return self.__time

    @property
    def item_path(self) -> str:
        """Python path like string to the object.

        :return: An import path
        :rtype: str
        """
        return self.__item_path

    @property
    def path(self) -> str:
        """Retrieves the test path relatively to current working directory where pytest has run.

        :return: A path
        :rtype: str
        """
        return self.__path

    @property
    def item(self) -> str:
        """Retrieves the test item name.
        If you want the test identifier, you should rather consider variant.
        Note that this might not be sufficient if multiple components have items named the same way.

        :return: The test item.
        :rtype: str
        """
        return self.__item

    @property
    def variant(self) -> str:
        """Retrieves the fully qualified test name, including parameter values if any.

        :return: A test identifier
        :rtype: str
        """
        return self.__variant

    @property
    def scope(self) -> MetricScope:
        """Retrieves the metrics scope (function, module or package).

        :return: the test scope
        :rtype: `monitor_api.MetricScope`
        """
        return self.__kind

    @property
    def component(self) -> str:
        """Retrieves the test component. Components can be viewed as a logical system representing
        an application or a subsystem.

        :return: the test's component name.
        :rtype: str
        """
        return self.__component

    @property
    def wall_time(self) -> float:
        """Retrieves the total, real time spent running the test.
        The unit is the second.

        :return: Time spent running the test.
        :rtype: float
        """
        return self.__wall

    @property
    def user_time(self) -> float:
        """Retrieves the amount of time spent running the test in user mode.
        The unit is the second.

        :return: Time spent running in user mode.
        :rtype: float
        """
        return self.__user

    @property
    def kernel_time(self) -> float:
        """Retrieves the amount of time spent running in kernel mode.
        The unit is the second.

        :return: Time spent running in kernel mode.
        :rtype: float
        """
        return self.__kernel

    @property
    def cpu_usage(self) -> float:
        """Retrieves the cpu usage (in percentage) of the test.

        Example
          -  0.6577 means 65.77 % use of a single CPU.
          -  3.768 means 3 CPU at 100 % plus another one at 76.8 %.

        :return: The test's cpu usage.
        :rtype: float
        """
        return self.__cpu

    @property
    def memory_usage(self) -> float:
        """Retrieves the amount of memory used by the test.
        The unit is the MB.

        :return: Memory used (in MB)
        :rtype: float
        """
        return self.__memory

    def to_dict(self, keep: List[Field] = None, drop: List[Field] = None) -> dict:
        """Convert the instance to a dictionary for further use. You can control which field
        will be exported using a keep/drop approach. Note that controlling the export can only be
        done using either `keep` or `drop` parameter.

        :param keep: List of field to keep. Use it if the field you want to focus on is small.
        :param drop: List of field to drop. Use it if the field you want to focus on is long.
        :type keep: list(`monitor_api.Field`)
        :type drop: list(`monitor_api.Field`)
        :return: A dictionary containing members whitelisted for export.
        :rtype: dict
        """
        d = {Field.CONTEXT_H.value: self.context,
             Field.SESSION_H.value: self.session,
             Field.ITEM_START_TIME.value: self.run_time,
             Field.ITEM_PATH.value: self.item_path,
             Field.ITEM.value: self.item,
             Field.ITEM_FS_LOC.value: self.path,
             Field.ITEM_VARIANT.value: self.variant,
             Field.KIND.value: self.scope.value,
             Field.TOTAL_TIME.value: self.wall_time,
             Field.USER_TIME.value: self.user_time,
             Field.KERNEL_TIME.value: self.kernel_time,
             Field.CPU_USAGE.value: self.cpu_usage,
             Field.MEM_USAGE.value: self.memory_usage,
             Field.COMPONENT.value: self.component}

        if keep is not None:
            cols = [col for col in METRIC_ALL_FIELDS if col not in keep]
        elif drop is not None:
            cols = [col for col in METRIC_ALL_FIELDS if col in drop]
        else:
            cols = []

        for col in cols:
            del d[col.value]
        return d

    def hash(self):
        """ Emulate hash key to avoid the same metric to be inserted multiple times
        in Metrics resource (useful for some entry points).

        :return: A hash key that uniquely identifies the test.
        :rtype: str
        """
        hr = hashlib.sha256()
        hr.update(self.context.encode())
        hr.update(self.session.encode())
        hr.update(self.run_time.isoformat().encode())
        hr.update(self.item_path.encode())
        hr.update(self.item.encode())
        hr.update(self.path.encode())
        hr.update(self.variant.encode())
        hr.update(self.scope.name.encode())
        hr.update(self.component.encode())
        hr.update(str(self.wall_time).encode())
        hr.update(str(self.user_time).encode())
        hr.update(str(self.kernel_time).encode())
        hr.update(str(self.cpu_usage).encode())
        hr.update(str(self.memory_usage).encode())
        return hr.hexdigest()


class Context:
    """Single execution context record. This is basically the machine which executed tests.

    :param h: Context identifier.
    :type h: str
    :param cpu_count: Number of CPU on the machine
    :type cpu_count: int
    :param cpu_freq: Nominal frequency of the CPI
    :type cpu_freq: int
    :param cpu_type: Architecture of the CPU
    :type cpu_type: str
    :param cpu_vendor: The CPU Vendor. Optional.
    :type cpu_vendor: str
    :param ram_total: Total amount of physical RAM available on the machine.
    :type ram_total: int
    :param mac_node: Fully qualified domain name of the machine.
    :type mac_node: str
    :param mac_type: Machine/OS type
    :type mac_type: str
    :param mac_arch: Machine architecture (32 or 64bits)
    :type mac_arch: str
    :param sys_info: OS kernel information.
    :type sys_info: str
    :param py_info: Python information.
    :type py_info: str
    """
    def __init__(self, h: str = None, cpu_count: int = None, cpu_freq: int = None, cpu_type: str = None,
                 cpu_vendor: str = None, ram_total: int = None, mac_node: str = None, mac_type: str = None,
                 mac_arch: str = None, sys_info: str = None, py_info: str = None):
        self.__h = h or ''
        self.__cpu_count = cpu_count or 1
        self.__cpu_freq = cpu_freq or 0
        self.__cpu_type = cpu_type or ''
        self.__cpu_vendor = cpu_vendor or ''
        self.__ram = ram_total or 0
        self.__mac_node = mac_node or ''
        self.__mac_type = mac_type or ''
        self.__mac_arch = mac_arch or ''
        self.__sys = sys_info or ''
        self.__py = py_info or ''

    @property
    def h(self) -> str:
        """Context identifier. Unique across all contexts. If two identifiers share the same
        value, it means that they point to objects with similar values.

        :return: A unique key for referencing this context.
        :rtype: str
        """
        return self.__h

    @property
    def cpu_count(self) -> int:
        """Total number of CPU on the machine.

        :return: The number of CPU on the machine
        :rtype: int
        """
        return self.__cpu_count

    @property
    def cpu_freq(self) -> int:
        """Nominal CPU frequency on the machine. Expressed in MHz

        :return: The CPU frequency expressed in MHz.
        :rtype: int
        """
        return self.__cpu_freq

    @property
    def cpu_type(self) -> str:
        """Architecture of the CPU.

        :return: Architecture identifier string.
        :rtype: str
        """
        return self.__cpu_type

    @property
    def cpu_vendor(self) -> str:
        """CPU model name, as provided by vendor.

        :return: CPU brand name
        :rtype: str
        """
        return self.__cpu_vendor

    @property
    def total_ram(self) -> int:
        """Total amount of RAM available on the machine, expressed in MB.

        :return: The amount of RAM (in MB)
        :rtype: int
        """
        return self.__ram

    @property
    def machine_node(self) -> str:
        """The machine fully qualified domain name.

        :return: the machine hostname.
        :rtype: str
        """
        return self.__mac_node

    @property
    def machine_type(self) -> str:
        """The machine type details the instruction set supported by the processor.

        :return: the machine type.
        :rtype: str
        """
        return self.__mac_type

    @property
    def machine_arch(self) -> str:
        """The machine architecture (32 or 64 bits).

        :return: the machine bitmode.
        :rtype: str
        """
        return self.__mac_arch

    @property
    def sys_info(self) -> str:
        """A string detailing some information about the OS kernel.
        Can explain the exact version of the kernel.

        :return: A informative string about the OS kernel.
        :rtype: str
        """
        return self.__sys

    @property
    def python_info(self) -> str:
        """A string detailing some information about the version of Python used for testing.
        Can explain how the interpreter has been compiled, on which system and when.

        :return: A informative string about the python interpreter.
        :rtype: str
        """
        return self.__py

    def to_dict(self, keep: List[Field] = None, drop: List[Field] = None) -> dict:
        """Convert the instance to a dictionary for further use. You can control which field
        will be exported using a keep/drop approach. Note that controlling the export can only be
        done using either `keep` or `drop` parameter.

        :param keep: List of field to keep. Use it if the field you want to focus on is small.
        :param drop: List of field to drop. Use it if the field you want to focus on is long.
        :type keep: list(`monitor_api.Field`)
        :type drop: list(`monitor_api.Field`)
        :return: A dictionary containing members whitelisted for export
        :rtype: dict
        """
        d = {Field.CONTEXT_H.value: self.h,
             Field.CPU_COUNT.value: self.cpu_count,
             Field.CPU_FREQUENCY_MHZ.value: self.cpu_freq,
             Field.CPU_TYPE.value: self.cpu_type,
             Field.CPU_VENDOR.value: self.cpu_vendor,
             Field.MACHINE_ARCH.value: self.machine_arch,
             Field.MACHINE_NODE.value: self.machine_node,
             Field.MACHINE_TYPE.value: self.machine_type,
             Field.RAM_TOTAL_MB.value: self.total_ram,
             Field.SYSTEM_INFO.value: self.sys_info,
             Field.PYTHON_INFO.value: self.python_info}

        if keep is not None:
            cols = [col for col in CONTEXT_ALL_FIELDS if col not in keep]
        elif drop is not None:
            cols = [col for col in CONTEXT_ALL_FIELDS if col in drop]
        else:
            cols = []

        for col in cols:
            del d[col.value]
        return d


class Session:
    """A session record. Contains SCM reference, start date and tags.

    :param h: session identifier
    :type h: str
    :param scm_ref: Source code management reference
    :type scm_ref: str
    :param run_date: Date time at which the session has been started.
    :type run_date: str
    :param tags: Additional tags used for identifying special run conditions.
    :type tags: dict
    """
    def __init__(self, h: str = None, scm_ref: str = None,
                 run_date: str = None,
                 tags: Optional[Union[Dict[str, str], List[Dict[str, str]]]] = None):
        self.__h = h or ''
        self.__scm = scm_ref or ''
        self.__start_date = api_fromisoformat(run_date)
        self.__tags = tags or dict()
        if type(tags) is list:
            # We try to map to a session sent by remote server
            try:
                d = dict()
                for item in tags:
                    d[item['name']] = item['value']
                self.__tags = d
            except KeyError:
                self.__tags = dict()

    def __str__(self):
        tags = None
        if self.__tags:
            tags = [f'        - {k}: {v}' for k, v in self.__tags.items()]
            tags = '\n'.join(tags)
        if tags is None:
            return f'''{self.__h}:
    run_date: {self.__start_date}
    scm: {self.__scm}
'''
        else:
            return f'''{self.__h}:
    run_date: {self.__start_date}
    scm: {self.__scm}
    tags:
{tags}
'''

    @property
    def tags(self) -> dict:
        """Retrieves all tags set on this session.
        All tags have their values represented as string.

        :return: A dictionary with key as tags and values as tag's value.
        :rtype: dict[str, str]
        """
        return self.__tags

    @property
    def h(self) -> str:
        """Session identifier. Unique across all sessions. If two identifiers share the same
        value, it means that they point to objects with similar values.

        :return: A unique key for referencing this session.
        :rtype: str
        """
        return self.__h

    @property
    def scm(self) -> str:
        """SCM reference to the code which have been executed.

        :return: A SCM reference
        :rtype: str"""
        return self.__scm

    @property
    def start_date(self) -> datetime.datetime:
        """Retrieve the start date at which the test session has started.

        :return: A datetime object positioned at the time the session started.
        :rtype: datetime.datetime
        """
        return self.__start_date

    def to_dict(self, keep: List[Field] = None, drop: List[Field] = None,
                expand_tags: bool = False) -> dict:
        """Convert the instance to a dictionary for further use. You can control which field
        will be exported using a keep/drop approach. Tags can be exported using an expand strategy
        and if requested, each tag is exported as if it was a real session field. This means that
        the final dictionary will have a depth of 1.
        Otherwise, tags will be grouped under a 'tags' element thus giving access to a sub dictionary
        containing all tags and only all tags. In this case, the final dictionary will have a depth of 2.

        Note that controlling the export can only be done using either `keep` or `drop` parameter.

        :param keep: List of field to keep. Use it if the field you want to focus on is small.
        :type keep: list(`monitor_api.Field`)
        :param drop: List of field to drop. Use it if the field you want to focus on is long.
        :type drop: list(`monitor_api.Field`)
        :param expand_tags: control how tags are exported. If set to True, tags will be exported like other
                            session members.
        :type expand_tags: bool
        :return: A dictionary containing members whitelisted for export
        :rtype: dict
        """
        if not expand_tags:
            d = {Field.SESSION_H.value: self.h,
                 Field.RUN_DATE.value: self.start_date,
                 Field.SCM.value: self.scm,
                 Field.TAGS.value: self.tags}
        else:
            d = {Field.SESSION_H.value: self.h,
                 Field.RUN_DATE.value: self.start_date,
                 Field.SCM.value: self.scm}
            d.update(self.tags)

        # If expand_tags, we keep (resp. drop) all tag fields
        dk_tags = list(set(d.keys()).difference(set([f.value for f in SESSION_ALL_FIELDS])))
        cols = []
        if keep is not None:
            cols = [col.value for col in SESSION_ALL_FIELDS if col not in keep]
            if expand_tags and Field.TAGS.value in cols:
                cols.remove(Field.TAGS.value)
                cols.extend(dk_tags)
        elif drop is not None:
            cols = [col.value for col in SESSION_ALL_FIELDS if col in drop]
            if expand_tags and Field.TAGS.value in cols:
                cols.remove(Field.TAGS.value)
                cols.extend(dk_tags)

        for col in cols:
            del d[col]
        return d



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT
#
# import enum
#
#
# class MetricType(enum.Enum):
#     FUNCTION = 'function'
#     MODULE = 'module'
#     PACKAGE = 'package'
#
#
# class MetricField(enum.Enum):
#     SESSION_H = 'session_h'
#     CONTEXT_H = 'context_h'
#     ITEM_START_TIME = 'start_time'
#     ITEM_PATH = 'item_path'
#     ITEM = 'item'
#     ITEM_VARIANT = 'variant'
#     ITEM_FS_LOC = 'path'
#     KIND = 'kind'
#     COMPONENT = 'component'
#     TOTAL_TIME = 'wall_time'
#     USER_TIME = 'user_time'
#     KERNEL_TIME = 'kernel_time'
#     CPU_USAGE = 'cpu_usage'
#     MEM_USAGE = 'memory_usage'
#
#
# METRIC_ALL_FIELDS = [MetricField.CONTEXT_H, MetricField.SESSION_H, MetricField.ITEM_START_TIME,
#                      MetricField.ITEM_PATH, MetricField.ITEM, MetricField.ITEM_FS_LOC,
#                      MetricField.ITEM_VARIANT, MetricField.KIND, MetricField.TOTAL_TIME,
#                      MetricField.USER_TIME, MetricField.KERNEL_TIME, MetricField.CPU_USAGE,
#                      MetricField.MEM_USAGE, MetricField.COMPONENT]
#
#
# class SessionField(enum.Enum):
#     H = 'h'
#     SCM = 'scm'
#     RUN_DATE = 'run_date'
#     TAGS = 'tags'
#
#
# SESSION_ALL_FIELDS = [SessionField.H, SessionField.SCM, SessionField.RUN_DATE, SessionField.TAGS]
#
#
# class ContextField(enum.Enum):
#     H = 'h'
#     CPU_COUNT = 'cpu_count'
#     CPU_FREQUENCY_MHZ = 'cpu_freq'
#     CPU_TYPE = 'cpu_type'
#     CPU_VENDOR = 'cpu_vendor'
#     RAM_TOTAL_MB = 'ram'
#     MACHINE_NODE = 'hostname'
#     MACHINE_TYPE = 'type'
#     MACHINE_ARCH = 'arch'
#     SYSTEM_INFO = 'sys'
#     PYTHON_INFO = 'py'
#
#
# CONTEXT_ALL_FIELDS = [ContextField.H, ContextField.CPU_COUNT, ContextField.CPU_FREQUENCY_MHZ,
#                       ContextField.CPU_TYPE, ContextField.CPU_VENDOR, ContextField.RAM_TOTAL_MB,
#                       ContextField.MACHINE_NODE, ContextField.MACHINE_TYPE, ContextField.MACHINE_ARCH,
#                       ContextField.SYSTEM_INFO, ContextField.PYTHON_INFO]



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from monitor_api.enums import MetricScope, ResourceType, ResourceMethod
from monitor_api.core.collections import Metrics, Sessions, Contexts
from monitor_api.core.entities import Session, Context
from monitor_api.dialects.local import Local
from monitor_api.dialects.remote import Remote
import typing


class Monitor:
    """Interface for manipulating your collected metrics either from a local storage or a remote server.

    :param str url: The url to the server, or the path to the local database.
    :Keyword Arguments for Remote server:
        Please refer to the `requests's Session <https://docs.python-requests.org/en/latest/api/#request-sessions>`_ .
        for more details

    :Keyword Arguments for Local server:
        Please refer to `sqlite3 connections <https://docs.python.org/3/library/sqlite3.html#connection-objects>`_
        for more details
    """
    def __init__(self, url: str, **kwargs):
        """Construct a new interface with a monitor storage.
        """
        if url.startswith('http://') or url.startswith('https://'):
            self.__dial = Remote(url, **kwargs)
        else:
            self.__dial = Local(url, **kwargs)

    def count_components(self) -> int:
        """Count total number of non empty components.

        :return: The number of non empty components
        :rtype: int
        """
        return self.__dial.count_components()

    def list_components(self) -> typing.List[str]:
        """Read all non empty components.

        :return: a list of non empty components
        :rtype: list(str)
        """
        return self.__dial.get_components()

    def list_component_pipelines(self, component: str) -> typing.List[str]:
        """Read for known pipelines on a given component.

        :param str component: Component for which this query must be done
        :return: The list of pipelines affiliated with this component
        :rtype: list(str)
        """
        return self.__dial.get_component_pipelines(component)

    def list_component_pipeline_builds(self, component: str, pipeline: str) -> typing.List[str]:
        """Read for known build ids of a pipeline on a given component.
        This is handy if you want to filter out sessions for comparison purpose.

        :param str component: Component for which this query must be done
        :param str pipeline: Pipeline identifier
        :return: The list of build ids affiliated with the pair pipeline/component
        :rtype: list(str)
        """
        return self.__dial.get_component_pipelines_build(component, pipeline)

    def list_component_metrics(self, component: typing.Optional[str] = None) -> Metrics:
        """Read for all metrics refering to the given component.

        :param str component: Optional. Component name. If None, metrics without component will be searched for.
        :return: A list of metrics, all of them having the requested component.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.read_component_metrics(component)

    def count_contexts(self) -> int:
        """Count the number of context entries.

        :return: The number of different context entries
        :rtype: int
        """
        return self.__dial.count_contexts()

    def list_contexts(self) -> Contexts:
        """Read all contexts and extract them into a Dict like structure for further preprocessing.
        The reference key is their hash member.

        :return: All contexts object referenced by their hash value.
        :rtype: monitor_api.core.collections.Contexts
        """
        return self.__dial.get_contexts()

    def get_context(self, context_h: str) -> typing.Optional[Context]:
        """Read and extract a single context object from a hash value.

        :param context_h: hash value of the context you wish to retrieve.
        :type context_h: str
        :return: A context object if there is a match between your hash key and accessed set of context,
                 None otherwise.
        :rtype: None or monitor_api.core.entities.Context
        """
        return self.__dial.get_context_details(context_h)

    def list_context_metrics(self, context_h: str) -> Metrics:
        """Read and extract all metrics having the given context hash value.

        :param context_h: hash value of the context you wish to retrieve.
        :type context_h: str
        :return: A list of metrics, all of which having their context key equal to the given context hash value.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_metrics_with_context(context_h)

    def count_metrics(self, session_h: str = None, context_h: str = None, scm_ref: str = None) -> int:
        """Count the number of metrics under some conditions.
        If no condition is set, the total number of metrics is computed.
        If session_h is set, count all metrics having session_h as session reference.
        Otherwise, if context_h is set, count all metrics having context_h as context reference.
        Otherwise, if scm_ref is set, count all metrics having a session pointing to this scm reference.

        :param session_h: A full session reference.
        :type session_h: str
        :param context_h: A full context reference.
        :type context_h: str
        :param scm_ref: A full scm reference.
        :type scm_ref: str
        :return: A number greater of equal to 0, -1 in case of access error.
        :rtype: int
        """
        return self.__dial.count_metrics(session_h, context_h, scm_ref)

    def list_item_metrics(self, item: str) -> Metrics:
        """Read all metrics for the given test item, mixing variant metrics together if your test is parameterized.

        :param item: the name of the test for which metrics must be retrieved.
        :rtype item: str
        :return: A (possibly empty) list of metrics.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_item_metrics(item)

    def list_metrics(self) -> Metrics:
        """Read all metrics.

        :return: A (possibly empty) list of metrics.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_metrics()

    def list_metrics_by_type(self, metric_type: MetricScope) -> Metrics:
        """Read all metrics linked to the given type of metrics.

        :param metric_type: the type of the test entity (function, module or package).
        :rtype metric_type: MetricType
        :return: A (possibly empty) list of metrics.
        :rtype: monitor_api.core.collections.Metrics
        :note: metric_type value package is not supported for now.
        """
        return self.__dial.get_metrics_by_type(metric_type)

    def list_metrics_by_scm_id(self, scm_ref: str) -> Metrics:
        """Read all metrics having a session pointing to this scm reference.

        :param scm_ref: Your entire scm reference
        :rtype scm_ref: str
        :return: A (possibly empty) list of metrics.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_metrics_with_scm_ref(scm_ref)

    def list_metrics_from_pattern(self, item: str = None, variant: str = None) -> Metrics:
        """Read all metrics using a prefix expression on either the test function or its variant.

        :param item: Prefix of your item. Leave it None if you want to focus on variants.
        :rtype item: str
        :param variant: Prefix of your variant. Leave it None if you want to focus on item.
        :rtype variant: str
        :return: A (possibly empty) list of metrics.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_metrics_by_pattern(item, variant)

    def list_metrics_of_variant(self, variant: str, component: str = None) -> Metrics:
        """Read all metrics for the given test variant, with ability to restrict the view with the component.

        :param variant: the name of the test for which metrics must be retrieved.
        :rtype variant: str
        :param component: Component name. Enables component restriction if set.
        :rtype component: str
        :return: A (possibly empty) list of metrics.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_metrics_of_variant(variant, component)

    def count_sessions(self) -> int:
        """Count the number of session entries.

        :return: The number of different session entries
        :rtype: int
        """
        return self.__dial.count_sessions()

    def list_sessions(self, with_tags: typing.Union[str, typing.List[str]] = None,
                      restrict_flags: typing.Union[str, typing.List[str]] = None,
                      method: str = None) -> Sessions:
        """Read all sessions matching your criteria. If no filters are provided, all sessions are retrieved.

        :param with_tags: Set the tags you want your sessions must have.
        :type with_tags: either str or list(str)
        :param restrict_flags: restrict you tag's values. Must have same length has with_tags.
        :type restrict_flags: either str or list(str)
        :param method: indicates whether you want a strict match (match_all) or a permissive one (match_any)
        :type method: str (allowed values are match_all and match_any)
        :return: The list of sessions matching the conditions.
        :rtype: monitor_api.core.collections.Sessions
        :note: This interface is subject for change in future version!
        """
        return self.__dial.get_sessions(with_tags, restrict_flags, method)

    def list_session_metrics(self, session_h: str = None) -> Metrics:
        """Read all metrics having the given session reference.

        :param session_h: The full session reference
        :type session_h: str
        :return: A (possibly empty) list of metrics.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_metrics_by_session(session_h)

    def get_session(self, session_h: str) -> typing.Optional[Session]:
        """Retrieve the session object having the given session reference.

        :param session_h: The full session reference
        :type session_h: str
        :return: A session object if the provided reference has a match, None otherwise.
        :rtype: None or monitor_api.core.entities.Session
        """
        return self.__dial.get_session_details(session_h)

    def list_pipelines(self) -> typing.List[str]:
        """Retrieve the list of known pipelines

        :return: the list of pipeline names
        :rtype: list(str)
        """
        return self.__dial.get_pipelines()

    def list_pipeline_builds(self, pipeline: str) -> typing.List[str]:
        """For a given pipeline, list all known build identifiers.

        :param pipeline: The pipeline for which all build must be listed.
        :type pipeline: str
        :return: The list of builds associated to this pipeline
        :rtype: list(str)
        """
        return self.__dial.get_pipeline_builds(pipeline)

    def list_metrics_resources(self, resource: ResourceType, method: ResourceMethod,
                               max_element: int = 10) -> Metrics:
        """Retrieve metrics having the higher (resp. lower) resource consumption accross the whole sets of metrics.
        Operates on at most 500 entries.

        :param resource: The resource you want to operate the extraction on.
        :type resource: ResourceType
        :param method: Indicate if you want greediest or most disciplined tests.
        :type method: ResourceType
        :param max_element: Set the number of element you wish to retrieve. Default is 10.
        :type max_element: int
        :return: A (possibly empty) list of metrics.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_metrics_by_resource(resource, method, max_element)

    def list_metrics_resources_from_component(self, component: str, resource: ResourceType,
                                              method: ResourceMethod, max_element: int = 10) -> Metrics:
        """Retrieve metrics having the higher (resp. lower) resource consumption for a given component.
        Operates on at most 500 entries.

        :param component: A valid component identifier
        :type component: str
        :param resource: The resource you want to operate the extraction on.
        :type resource: ResourceType
        :param method: Indicate if you want greediest or most disciplined tests.
        :type method: ResourceType
        :param max_element: Set the number of element you wish to retrieve. Default is 10.
        :type max_element: int
        :return: A (possibly empty) list of metrics.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_metrics_by_component_resource(resource, component, method, max_element)

    def list_metrics_resources_from_pipeline(self, pipeline: str, resource: ResourceType,
                                             method: ResourceMethod, max_element: int = 10) -> Metrics:
        """Retrieve metrics having the higher (resp. lower) resource consumption accross a pipeline (multiple builds).
        Operates on at most 500 entries.

        :param pipeline: The pipeline name of the build.
        :type pipeline: str
        :param resource: The resource you want to operate the extraction on.
        :type resource: ResourceType
        :param method: Indicate if you want greediest or most disciplined tests.
        :type method: ResourceType
        :param max_element: Set the number of element you wish to retrieve. Default is 10.
        :type max_element: int
        :return: A (possibly empty) list of metrics.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_metrics_by_pipeline_resource(resource, pipeline, method, max_element)

    def list_metrics_resources_from_build(self, pipeline: str, build: str, resource: ResourceType,
                                          method: ResourceMethod, max_element: int = 10) -> Metrics:
        """Retrieve metrics having the higher (resp. lower) resource consumption for a full build reference.
        Operates on at most 500 entries.

        :param pipeline: The pipeline name of the build.
        :type pipeline: str
        :param build: The build identifier
        :type build: str
        :param resource: The resource you want to operate the extraction on.
        :type resource: ResourceType
        :param method: Indicate if you want greediest or most disciplined tests.
        :type method: ResourceType
        :param max_element: Set the number of element you wish to retrieve. Default is 10.
        :type max_element: int
        :return: A (possibly empty) list of metrics.
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_metrics_by_build_resource(resource, pipeline, build, method, max_element)

    def list_metrics_from(self, sessions: Sessions = None, contexts: Context = None) -> Metrics:
        """For sessions and contexts, retrieve all metrics referencing at least a session or a context.

        :param sessions: Your set of sessions
        :type sessions: Sessions
        :param contexts: Your set of contexts
        :type contexts: Contexts
        :return: A (possibly empty) list of metrics
        :rtype: monitor_api.core.collections.Metrics
        """
        return self.__dial.get_metrics_from(sessions, contexts)

    def list_sessions_from(self, metrics: Metrics) -> Sessions:
        """Retrieve all sessions referenced in the metrics set.

        :param metrics: A set of metrics
        :type metrics: Metrics
        :return: All sessions object reference in your Metrics object.
        :rtype: monitor_api.core.collections.Sessions
        """
        return self.__dial.get_sessions_from(metrics)

    def list_contexts_from(self, metrics: Metrics) -> Contexts:
        """Retrieve all contexts referenced in the metrics set.

        :param metrics: A set of metrics
        :type metrics: Metrics
        :return: All context objects reference in your Metrics object.
        :rtype: monitor_api.core.collections.Contexts
        """
        return self.__dial.get_contexts_from(metrics)

    def list_build_sessions(self, pipeline: str, build_id: str) -> Sessions:
        """Retrieve all sessions for the given pipeline and build identifier.

        :param pipeline: The pipeline identifier
        :type pipeline: str
        :param build_id: The build identifier
        :type build_id: str
        :return: All session objects for the given build.
        :rtype: monitor_api.core.collections.Sessions
        """
        return self.__dial.get_sessions_from_build(pipeline, build_id)



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from flask_restx import Resource

from sqlalchemy import func

from http import HTTPStatus
from monitor_server import SERVER
from monitor_server.api.model import (resources_ns as ns, resource_metrics_list_mode)
from monitor_server.data.model import MetricModel, SessionModel


J_PIPELINE = '$.pipeline_branch'
J_BUILD = '$.pipeline_build_no'
RESOURCE_2_ENTITY = dict(memory=MetricModel.mem_usage, cpu=MetricModel.cpu_usage,
                         kernel_time=MetricModel.krnl_time, user_time=MetricModel.user_time,
                         total_time=MetricModel.wall_time)


@ns.route('/<string:resource>/head/metrics',
          defaults=dict(max_element=10),
          doc=dict(description='List up to 10 metrics from tests which consumes '
                               'most the given <resource>.'))
@ns.route('/<string:resource>/head/<int:max_element>/metrics',
          doc=dict(description='List up to <max_element> metrics from tests which consumes '
                               'most the given <resource>.'))
class ListHeadResource(Resource):
    @ns.marshal_list_with(resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.OK), 'Success', resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (No content)', resource_metrics_list_mode)
    def get(self, resource, max_element):
        max_element = min(max_element, 500)
        query = MetricModel.query.order_by(RESOURCE_2_ENTITY[resource].desc()).limit(max_element)
        items = [item for item in query]
        return dict(metrics=items)


@ns.route('/<string:resource>/tail/metrics',
          defaults=dict(max_element=10),
          doc=dict(description='List up to 10 metrics from tests which consumes '
                               'least the given <resource>.'))
@ns.route('/<string:resource>/tail/<int:max_element>/metrics',
          doc=dict(description='List up to <max_element> metrics from tests which consumes '
                               'least the given <resource>.'))
class ListTailResource(Resource):
    @ns.marshal_list_with(resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.OK), 'Success', resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (No content)', resource_metrics_list_mode)
    def get(self, resource, max_element):
        max_element = min(max_element, 500)
        query = MetricModel.query.order_by(RESOURCE_2_ENTITY[resource]).limit(max_element)
        items = [item for item in query]
        return dict(metrics=items)


# By pipeline
@ns.route('/<string:resource>/pipelines/<string:pipeline>/head/metrics',
          defaults=dict(max_element=10),
          doc=dict(description='List up to 10 metrics from tests which consumes'
                               'most the given <resource> on <pipeline>.'))
@ns.route('/<string:resource>/pipelines/<string:pipeline>/head/<int:max_element>/metrics',
          doc=dict(description='List up to <max_element> metrics from tests which consumes'
                               'most the given <resource> on <pipeline>.'))
class ListHeadResourceOfPipeline(Resource):
    @ns.marshal_list_with(resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.OK), 'Success', resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (No content)', resource_metrics_list_mode)
    def get(self, resource, pipeline, max_element):
        max_element = min(max_element, 500)
        session = SERVER.DB.session
        query = session.query(MetricModel)
        query = query.filter(func.json_extract(SessionModel.description, J_PIPELINE) == pipeline,
                             MetricModel.session_h == SessionModel.h)
        query = query.order_by(RESOURCE_2_ENTITY[resource].desc()).limit(max_element)
        items = [item for item in query]
        return dict(metrics=items)


@ns.route('/<string:resource>/pipelines/<string:pipeline>/tail/metrics',
          defaults=dict(max_element=10),
          doc=dict(description='List up to 10 metrics from tests which consumes'
                               'least the given <resource> on <pipeline>.'))
@ns.route('/<string:resource>/pipelines/<string:pipeline>/tail/<int:max_element>/metrics',
          doc=dict(description='List up to <max_element> metrics from tests which consumes'
                               'least the given <resource> on <pipeline>.'))
class ListTailResourceOfPipeline(Resource):
    @ns.marshal_list_with(resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.OK), 'Success', resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (No content)', resource_metrics_list_mode)
    def get(self, resource, pipeline, max_element):
        max_element = min(max_element, 500)
        session = SERVER.DB.session
        query = session.query(MetricModel)
        query = query.filter(func.json_extract(SessionModel.description, J_PIPELINE) == pipeline,
                             MetricModel.session_h == SessionModel.h)
        query = query.order_by(RESOURCE_2_ENTITY[resource]).limit(max_element)
        items = [item for item in query]
        return dict(metrics=items)


# By build
@ns.route('/<string:resource>/pipelines/<string:pipeline>/builds/<string:build>/head/metrics',
          defaults=dict(max_element=10),
          doc=dict(description='List up to 10 metrics from tests which consumes'
                               'most the given <resource> on (<pipeline>, <build>).'))
@ns.route('/<string:resource>/pipelines/<string:pipeline>/builds/<string:build>/head/<int:max_element>/metrics',
          doc=dict(description='List up to <max_element> metrics from tests which consumes'
                               'most the given <resource> on (<pipeline>, <build>).'))
class ListHeadResourceOfBuild(Resource):
    @ns.marshal_list_with(resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.OK), 'Success', resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (No content)', resource_metrics_list_mode)
    def get(self, resource, pipeline, build, max_element):
        max_element = min(max_element, 500)
        session = SERVER.DB.session
        query = session.query(MetricModel)
        query = query.filter(func.json_extract(SessionModel.description, J_PIPELINE) == pipeline,
                             func.json_extract(SessionModel.description, J_BUILD) == build,
                             MetricModel.session_h == SessionModel.h)
        query = query.order_by(RESOURCE_2_ENTITY[resource].desc()).limit(max_element)
        items = [item for item in query]
        return dict(metrics=items)


@ns.route('/<string:resource>/pipelines/<string:pipeline>/builds/<string:build>/tail/metrics',
          defaults=dict(max_element=10),
          doc=dict(description='List up to 10 metrics from tests which consumes'
                               'least the given <resource> on (<pipeline>, <build>).'))
@ns.route('/<string:resource>/pipelines/<string:pipeline>/builds/<string:build>/tail/<int:max_element>/metrics',
          doc=dict(description='List up to <max_element> metrics from tests which consumes'
                               'least the given <resource> on (<pipeline>, <build>).'))
class ListTailResourceOfBuild(Resource):
    @ns.marshal_list_with(resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.OK), 'Success', resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (No content)', resource_metrics_list_mode)
    def get(self, resource, pipeline, build, max_element):
        max_element = min(max_element, 500)
        session = SERVER.DB.session
        query = session.query(MetricModel)
        query = query.filter(func.json_extract(SessionModel.description, J_PIPELINE) == pipeline,
                             func.json_extract(SessionModel.description, J_BUILD) == build,
                             MetricModel.session_h == SessionModel.h)
        query = query.order_by(RESOURCE_2_ENTITY[resource]).limit(max_element)
        items = [item for item in query]
        return dict(metrics=items)


# By Component
@ns.route('/<string:resource>/components/<string:component>/head/metrics',
          defaults=dict(max_element=10),
          doc=dict(description='List up to 10 metrics from <component> which consumes '
                               'most the given <resource>.'))
@ns.route('/<string:resource>/components/<string:component>/head/<int:max_element>/metrics',
          doc=dict(description='List up to <max_element> metrics from <component> which consumes '
                               'most the given <resource>.'))
class ListHeadResourceOfComponent(Resource):
    @ns.marshal_list_with(resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.OK), 'Success', resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (No content)', resource_metrics_list_mode)
    def get(self, resource, component, max_element):
        max_element = min(max_element, 500)
        query = MetricModel.query.filter(MetricModel.component == component)
        query = query.order_by(RESOURCE_2_ENTITY[resource].desc()).limit(max_element)
        items = [item for item in query]
        return dict(metrics=items)


@ns.route('/<string:resource>/components/<string:component>/tail/metrics',
          defaults=dict(max_element=10),
          doc=dict(description='List up to 10 metrics from <component> which consumes '
                               'least the given <resource>.'))
@ns.route('/<string:resource>/components/<string:component>/tail/<int:max_element>/metrics',
          doc=dict(description='List up to <max_element> metrics from <component> which consumes '
                               'least the given <resource>.'))
class ListTailResourceOfComponent(Resource):
    @ns.marshal_list_with(resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.OK), 'Success', resource_metrics_list_mode)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (No content)', resource_metrics_list_mode)
    def get(self, resource, component, max_element):
        max_element = min(max_element, 500)
        query = MetricModel.query.filter(MetricModel.component == component)
        query = query.order_by(RESOURCE_2_ENTITY[resource]).limit(max_element)
        items = [item for item in query]
        return dict(metrics=items)



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from flask_restx import fields
from monitor_server import SERVER


def make_pageable(model):
    pageable = dict(model)
    pageable.update({'prev_url': fields.String(required=False,
                                               description='Url to previous page'),
                     'next_url': fields.String(required=False,
                                               description='Url to next page'),
                     'total_page': fields.Integer(required=False, default=1,
                                                  description='Total number of pages')})
    return pageable


# ------------------------------------------------------------------------------------------
#  Global namespaces
# ------------------------------------------------------------------------------------------
components_ns = SERVER.API_V1.namespace('components',
                                        description='Component based search and collections.')
filters_ns = SERVER.API_V1.namespace('filters', validate=True,
                                     description='Reads metrics from special filters.')
items_ns = SERVER.API_V1.namespace('items', validate=True,
                                   description='Item based search and collections of metrics.')
metrics_ns = SERVER.API_V1.namespace('metrics', validate=True,
                                     description='Access all metrics.')
pipelines_ns = SERVER.API_V1.namespace('pipelines',
                                       description='Address your metrics from your CI builds.')
resources_ns = SERVER.API_V1.namespace('resources',
                                       description='Find resource usage by items and variants.')
sessions_ns = SERVER.API_V1.namespace('sessions',
                                      description='Session based search and collection.')
variants_ns = SERVER.API_V1.namespace('variants', validate=True,
                                      description='Variants based search and collections of metrics.')
xcontexts_ns = SERVER.API_V1.namespace('contexts', description='Execution contexts based search and collection.')

# ------------------------------------------------------------------------------------------
#  Global schema
# ------------------------------------------------------------------------------------------
COMPONENT_LIST_SCHEMA = {
    'components': fields.List(fields.String(description="A component name"),
                              required=True, description="Full components list")
}

COMPONENT_PIPELINES_SCHEMA = {
    'component': fields.String(description="A component name", required=True),
    'pipelines': fields.List(fields.String(description="A pipeline identifier"),
                             required=True, description="List of pipelines for the given component.")
}

COMPONENT_PIPELINE_BUILDS_SCHEMA = {
    'component': fields.String(description="A component name", required=True),
    'pipeline': fields.String(description="A pipeline identifier", required=True),
    'builds': fields.List(fields.String(description='A build identifier.'),
                          required=True, description='List of builds for the given component on the given pipeline.')
}

COMPONENT_PIPELINE_BUILD_SESSIONS_SCHEMA = {
    'component': fields.String(description="A component name", required=True),
    'pipeline': fields.String(description="A pipeline identifier", required=True),
    'build': fields.String(description='A build identifier.', required=True),
    'sessions': fields.List(fields.String('A session identifier.'),
                            required=True,
                            description='List of all sessions associated to the given build, component and pipeline.')
}

EXEC_CONTEXT_FIELD_SCHEMA = {
    'h': fields.String(required=True, description="Unique Context identifier"),
    'cpu_count': fields.Integer(required=True, description='Number of available CPU'),
    'cpu_frequency': fields.Integer(required=True, attribute='cpu_freq',
                                    description='Nominal frequency of a CPU (in MHz)'),
    'cpu_type': fields.String(required=True, description='Architecture of the CPU'),
    'cpu_vendor': fields.String(required=False,
                                description='Information about constructor.'),
    'ram_total': fields.Integer(required=True,
                                description='Total RAM available'),
    'machine_node': fields.String(required=True, attribute='mac_node',
                                  description="Fully Qualified Domain Name (FQDN) of the machine  used for testing"),
    'machine_type': fields.String(required=True, attribute='mac_type',
                                  description="Machine architecture"),
    'machine_arch': fields.String(required=True, attribute='mac_arch',
                                  description="Bitmode of the machine"),
    'system_info': fields.String(required=True, attribute='sys_info',
                                 description="Short string about the Operating System"),
    'python_info': fields.String(required=True, attribute='py_info',
                                 description="Python information")
}


METRIC_FIELD_SCHEMA = {
    'session_h': fields.String(required=True, description='Session id on which the test has been run.'),
    'context_h': fields.String(required=True, attribute='ctx_h',
                               description='ExecutionContext id on which the test has been run.'),
    'item_start_time': fields.String(required=True, description='Effective start time of the test. Format: iso8601.'),
    'item_path': fields.String(required=True, description='Path of the item using Python import specification.'),
    'item': fields.String(required=True, description='Fully qualified name of the test (path/function_name)'),
    'item_variant': fields.String(required=True, description='Parametrization of the item.'),
    'item_fs_loc': fields.String(required=True,
                                 description="Relative path from pytest invocation directory to the item's module."),
    'kind': fields.String(required=True, description='Item type (class, function, module...)'),
    'component': fields.String(required=True, description='Component to which the test belong'),
    'total_time': fields.Float(required=True, attribute='wall_time',
                               description='Total time spent running the test (in seconds)'),
    'user_time':  fields.Float(required=True, description='Total time spent running in user space (in seconds)'),
    'kernel_time': fields.Float(required=True, attribute='krnl_time',
                                description='Total time spent running in kernel mode (in seconds)'),
    'cpu_usage': fields.Float(required=True, description='Percentage of CPU used while the test ran.'),
    'mem_usage': fields.Float(required=True, description='Memory used while running the test (in MB)')
}


TAG_FIELD_SCHEMA = {
    "name": fields.String(required=True, description='Tag name'),
    "value": fields.String(required=True, description="The tag's value")
}

PIPELINE_BUILDS_SCHEMA = {
    "pipeline": fields.String(required=True, description="CI Pipeline associated with following builds id."),
    "builds": fields.List(fields.String(required=True, description='Single build identifier'),
                          required=True, description='All available builds id for the given pipeline.')
}

PIPELINE_BUILD_SESSIONS_SCHEMA = {
    "pipeline": fields.String(required=True, description="CI Pipeline associated with following builds id."),
    "build": fields.String(required=True, description="Build identifier who has issued the following sessions id."),
    "sessions": fields.List(fields.String(required=True, description='Single Session identifier'),
                            required=True, description='All available sessions id for given pipeline and build id.')
}

# ------------------------------------------------------------------------------------------
#  Specific schema, might have some very specific models instantiated
# ------------------------------------------------------------------------------------------
SINGLE_COMPONENT_SCHEMA = {
    'name': fields.String(required=True, description="Full component name")
}

COMPONENT_METRICS_FIELD_SCHEMA = {
    'metrics': fields.List(fields.Nested(components_ns.model('metrics', METRIC_FIELD_SCHEMA))),
    'component': fields.String(required=True, description="Full component name")
}

COMPONENT_COUNT_SCHEMA = {
    'count': fields.Integer(required=True, description='Number of components')
}


EXECUTION_CONTEXT_METRICS_FIELD_SCHEMA = {
    'metrics': fields.List(fields.Nested(xcontexts_ns.model('metrics', METRIC_FIELD_SCHEMA)))
}

EXECUTION_CONTEXT_COUNT_SCHEMA = {
    'count': fields.Integer(required=True, description="Number of known execution contexts.")
}

METRIC_COUNT_SCHEMA = {
    'count': fields.Integer(required=True, description='Total count of metrics stored.')
}

SESSION_METRICS_FIELD_SCHEMA = {
    'metrics': fields.List(fields.Nested(sessions_ns.model('metrics', METRIC_FIELD_SCHEMA)))
}

SESSION_COUNT_SCHEMA = {
    'count': fields.Integer(required=True, description='Total count of sessions stored.')
}

single_tag_model = sessions_ns.model('SingleTag', TAG_FIELD_SCHEMA)

SESSION_FIELD_SCHEMA = {
    'session_h': fields.String(required=True, attribute='h',
                               description='Session id on which the test has been run.'),
    'run_date': fields.DateTime(dt_format='iso8601', required=True,
                                description='Date and time when test session has started. Format: iso8601.'),
    'scm_ref': fields.String(required=False, description='Reference to the SCM if any.'),
    'tags': fields.List(fields.Nested(single_tag_model),
                        required=True, description='List of all tags attached to a session.')
}

filter_single_tag_model = filters_ns.model('SingleTag', TAG_FIELD_SCHEMA)

FILTER_SESSION_FIELD_SCHEMA = {
    'session_h': fields.String(required=True, attribute='h',
                               description='Session id on which the test has been run.'),
    'run_date': fields.DateTime(dt_format='iso8601', required=True,
                                description='Date and time when test session has started. Format: iso8601.'),
    'scm_ref': fields.String(required=False, description='Reference to the SCM if any.'),
    'tags': fields.List(fields.Nested(filter_single_tag_model),
                        required=True, description='List of all tags attached to a session.')
}

SINGLE_BUILD_SCHEMA = {
    'id': fields.String(required=True, description='Pipeline identifier'),
    'sessions': fields.List(fields.String(required=True, description='Session hash identifier'),
                            required=True, description='List of sessions ran by this pipeline')
}

BUILD_PIPELINE_SCHEMA = {
    'pipeline': fields.String(required=True, description='Pipeline name'),
    'builds': fields.List(fields.Nested(pipelines_ns.model('SessionsByBuild', SINGLE_BUILD_SCHEMA)),
                          required=True, description='Builds issued from this pipeline')
}

PIPELINE_LIST_SCHEMA = {
    'pipelines': fields.List(fields.String(required=True, description='A pipeline name'),
                             required=True, description='List of pipelines known by the system')
}

# ------------------------------------------------------------------------------------------
#  Final models
# ------------------------------------------------------------------------------------------
# Components
single_component_model = components_ns.model('Component', SINGLE_COMPONENT_SCHEMA)

metrics_by_component_model = components_ns.model('MetricsByComponent', make_pageable(COMPONENT_METRICS_FIELD_SCHEMA))

component_list_model = components_ns.model('ComponentsList', make_pageable(COMPONENT_LIST_SCHEMA))

component_count_model = components_ns.model('ComponentsCount', COMPONENT_COUNT_SCHEMA)

component_pipelines_model = components_ns.model('ComponentPipelines', make_pageable(COMPONENT_PIPELINES_SCHEMA))

component_pipeline_builds_model = components_ns.model('ComponentPipelineBuilds',
                                                      make_pageable(COMPONENT_PIPELINE_BUILDS_SCHEMA))

component_pipeline_build_sessions_model = components_ns.model('ComponentPipelineBuildSessions',
                                                              make_pageable(COMPONENT_PIPELINE_BUILD_SESSIONS_SCHEMA))

# Execution contexts
xcontext_model = xcontexts_ns.model('ExecutionContext', EXEC_CONTEXT_FIELD_SCHEMA)

xcontext_list_model = xcontexts_ns.model('ExecutionContextsList',
                                         make_pageable({'contexts': fields.List(fields.Nested(xcontext_model))})
                                         )

metrics_by_xcontext_model = xcontexts_ns.model('MetricsByExecutionContext',
                                               make_pageable(EXECUTION_CONTEXT_METRICS_FIELD_SCHEMA)
                                               )
metrics_by_xcontext_count_model = xcontexts_ns.model('MetricsCount', METRIC_COUNT_SCHEMA)

xcontext_count_model = xcontexts_ns.model('ExecutionContextsCount', EXECUTION_CONTEXT_COUNT_SCHEMA)

# Filters
filtr_metrics_count_model = filters_ns.model('MetricsCount', METRIC_COUNT_SCHEMA)

filtr_metric_model = filters_ns.model('Metric', METRIC_FIELD_SCHEMA)

filtr_metric_list_model = filters_ns.model('MetricsList',
                                           make_pageable(
                                               {'metrics': fields.List(fields.Nested(filtr_metric_model))}
                                           ))

filtr_session_model = filters_ns.model('Session', FILTER_SESSION_FIELD_SCHEMA)
filtr_session_list_model = filters_ns.model('SessionList',
                                            make_pageable(
                                                {'sessions': fields.List(fields.Nested(filtr_session_model))}
                                            ))

# Items
item_metrics_count_model = metrics_ns.model('MetricsCount', METRIC_COUNT_SCHEMA)

item_metric_model = metrics_ns.model('Metric', METRIC_FIELD_SCHEMA)

item_metric_list_model = metrics_ns.model('MetricsList',
                                          make_pageable({'metrics': fields.List(fields.Nested(item_metric_model))})
                                          )

# Metrics
metric_count_model = metrics_ns.model('MetricsCount', METRIC_COUNT_SCHEMA)

metric_model = metrics_ns.model('Metric', METRIC_FIELD_SCHEMA)

metric_list_model = metrics_ns.model('MetricsList',
                                     make_pageable({'metrics': fields.List(fields.Nested(metric_model))})
                                     )

# Sessions
session_model = sessions_ns.model('Session', SESSION_FIELD_SCHEMA)

session_list_model = sessions_ns.model('SessionsList',
                                       make_pageable({'sessions': fields.List(fields.Nested(session_model))})
                                       )

metrics_by_session_list_model = sessions_ns.model('MetricsBySession', make_pageable(SESSION_METRICS_FIELD_SCHEMA))

session_count_model = sessions_ns.model('SessionCount', SESSION_COUNT_SCHEMA)

metrics_by_session_count_model = sessions_ns.model('MetricsCount', METRIC_COUNT_SCHEMA)

# Pipelines & builds
pipeline_model = pipelines_ns.model('SessionsByPipeline', make_pageable(BUILD_PIPELINE_SCHEMA))

pipeline_list_model = pipelines_ns.model('PipelineList', PIPELINE_LIST_SCHEMA)

pipeline_builds_model = pipelines_ns.model('PipelineBuilds', make_pageable(PIPELINE_BUILDS_SCHEMA))

pipeline_build_sessions_model = pipelines_ns.model('PipelineBuildSessions',
                                                   make_pageable(PIPELINE_BUILD_SESSIONS_SCHEMA))

# Resources
resource_metric_model = resources_ns.model('Metrics', METRIC_FIELD_SCHEMA)
resource_metrics_list_mode = resources_ns.model('MetricsByResource',
                                                make_pageable({
                                                    "metrics": fields.List(fields.Nested(resource_metric_model))
                                                })
                                                )

# Variants
vart_metrics_count_model = metrics_ns.model('MetricsCount', METRIC_COUNT_SCHEMA)

vart_metric_model = metrics_ns.model('Metric', METRIC_FIELD_SCHEMA)

vart_metric_list_model = metrics_ns.model('MetricsList',
                                          make_pageable({'metrics': fields.List(fields.Nested(vart_metric_model))})
                                          )



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from flask import request
from flask_restx import Resource
from sqlalchemy import func, and_, or_

from http import HTTPStatus
from monitor_server import SERVER
from monitor_server.api.model import (sessions_ns as ns,
                                      session_count_model, session_list_model, metrics_by_session_count_model,
                                      metrics_by_session_list_model, session_model)
from monitor_server.api.pager_utils import make_page
from monitor_server.data.model import SessionModel, MetricModel


def _session_description_to_tag_listing(description):
    listing = [{'name': key, 'value': value} for key, value in description.items()]
    return listing


@ns.route('/', defaults=dict(session_hash=None),
          doc=dict(description='List all sessions.'))
@ns.route('/<string:session_hash>',
          doc=dict(description='List all sessions with key starting with :session_hash.'))
class ListSessions(Resource):
    @ns.marshal_list_with(session_list_model, skip_none=True)
    @ns.response(int(HTTPStatus.OK), 'Success', session_list_model)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', session_list_model)
    def get(self, session_hash):

        def collect(items):
            d = []
            for session in items:
                a_session = dict(h=session.h, run_date=session.run_date,
                                 scm_ref=session.scm_ref,
                                 tags=_session_description_to_tag_listing(session.description))
                d.append(a_session)
            return d

        if session_hash is None:
            query = SessionModel.query
            collector = collect
        else:
            session_hash = f'{session_hash}%'
            query = SessionModel.query.filter(SessionModel.h.like(session_hash))
            collector = collect

        if request.args.get('with_flags', ''):
            method = request.args.get('method', 'match_any')
            with_flags = [f'$.{i}' for i in request.args.get('with_flags').split(',')]
            flags_restriction = request.args.get('restrict_flags', '').split(',')
            if len(flags_restriction) < len(with_flags):
                flags_restriction += [''] * (len(with_flags) - len(flags_restriction))
            cond = []
            for restrict_flag, restrict_value in zip(with_flags, flags_restriction):
                if restrict_value:
                    cond.append(func.json_extract(SessionModel.description, restrict_flag) == restrict_value)
                else:
                    cond.append(func.json_extract(SessionModel.description, restrict_flag) != '')
            else:
                if method == 'match_any':
                    query = query.filter(or_(*cond))
                else:
                    query = query.filter(and_(*cond))

        return make_page(request, query, 'sessions', collector=collector)

    @ns.marshal_list_with(session_model)
    @ns.response(int(HTTPStatus.CREATED), 'Success', session_model)
    @ns.response(int(HTTPStatus.BAD_REQUEST), 'Error')
    def post(self, session_hash):
        if type(request.json['description']) is not dict:
            return dict(message='Bad description format'), HTTPStatus.BAD_REQUEST
        session = SessionModel(h=request.json.get('session_h', session_hash),
                               run_date=request.json['run_date'],
                               scm_ref=request.json['scm_ref'],
                               description=request.json['description'])

        SERVER.DB.session.add(session)
        SERVER.DB.session.commit()
        return dict(h=session.h, run_date=session.run_date,
                    scm_ref=session.scm_ref,
                    tags=_session_description_to_tag_listing(session.description)), HTTPStatus.CREATED


@ns.route('/count',
          doc=dict(description='Count all sessions.'))
class CountSessions(Resource):
    @ns.marshal_list_with(session_count_model)
    @ns.response(int(HTTPStatus.OK), 'Success', session_count_model)
    def get(self):
        db_session = SERVER.DB.session
        count = db_session.query(func.count(SessionModel.scm_ref))
        return dict(count=count.first()[0]), HTTPStatus.OK


@ns.route('/<string:session_hash>/metrics',
          doc=dict(description='List all metric related to session :session_hash.'))
class ListMetricsBySession(Resource):
    @ns.marshal_list_with(metrics_by_session_list_model, skip_none=True)
    @ns.response(int(HTTPStatus.OK), 'Success', metrics_by_session_list_model)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', metrics_by_session_list_model)
    def get(self, session_hash):
        query = MetricModel.query.filter(MetricModel.session_h == session_hash)
        return make_page(request, query, 'metrics')


@ns.route('/<string:session_hash>/metrics/count',
          doc=dict(description='Count all metric related to session :session_hash'))
class CountMetricsBySession(Resource):
    @ns.marshal_list_with(metrics_by_session_count_model, skip_none=True)
    @ns.response(int(HTTPStatus.OK), 'Success', metrics_by_session_count_model)
    def get(self, session_hash):
        db_session = SERVER.DB.session
        count = db_session.query(func.count(MetricModel.test_id))
        count = count.filter(MetricModel.session_h == session_hash)
        return dict(count=count.first()[0]), HTTPStatus.OK





# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from monitor_api.enums import ResourceType, ResourceMethod, MetricScope, Field
from monitor_api.monitor import Monitor
from monitor_api.core.collections import Sessions, Metrics, Contexts
from monitor_api.core.entities import Session, Metric, Context


__all__ = ["Monitor", "ResourceType", "ResourceMethod", "MetricScope", "Field",
           "Sessions", "Metrics", "Contexts", "Session", "Metric", "Context"]



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from flask import request
from flask_restx import Resource

from sqlalchemy import func, distinct

from http import HTTPStatus
from monitor_server import SERVER
from monitor_server.api.model import (components_ns as ns, component_pipelines_model,
                                      component_pipeline_builds_model, component_pipeline_build_sessions_model,
                                      component_list_model, component_count_model, metrics_by_component_model)
from monitor_server.api.pager_utils import make_page
from monitor_server.data.model import MetricModel, SessionModel


@ns.route('/', defaults=dict(component=None),
          doc=dict(description='List all components.'))
@ns.route('/<string:component>',
          doc=dict(description='List all components having :component as prefix'))
class ListComponents(Resource):
    @ns.marshal_list_with(component_list_model, skip_none=True)
    @ns.response(int(HTTPStatus.OK), 'Success', component_list_model)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', component_list_model)
    def get(self, component):
        all_components = SERVER.DB.session.query(MetricModel.component)
        if component is not None:
            pattern = f'{component}%'
            all_components = all_components.filter(MetricModel.component.like(pattern))
        all_components = all_components.distinct()
        return make_page(request, all_components, 'components', lambda x: [i[0] for i in x])


@ns.route('/count',
          doc=dict(description='Count total number of components (empty components are not taken into account.'))
class CountComponents(Resource):
    @ns.marshal_list_with(component_count_model)
    @ns.response(int(HTTPStatus.OK), 'Success', component_count_model)
    def get(self):
        count = SERVER.DB.session.query(func.count(distinct(MetricModel.component)))
        count = count.first()[0]
        return dict(count=max(count, count - 1)), HTTPStatus.OK


@ns.route('/<string:component>/metrics',
          doc=dict(description='List all metrics having belonging to component :component'))
@ns.route('/metrics', defaults=dict(component=None),
          doc=dict(description='List all metrics not belonging to a component (component is empty).'))
class ListMetricsByComponent(Resource):
    @ns.marshal_list_with(metrics_by_component_model, skip_none=True)
    @ns.response(int(HTTPStatus.OK), 'Success', metrics_by_component_model)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)')
    def get(self, component):
        if component is None:
            component = ''
        return make_page(request,
                         MetricModel.query.filter(MetricModel.component.like(component)),
                         'metrics')


@ns.route('/<string:component>/pipelines',
          doc=dict(description='List all pipelines which issued metrics for the component :component'))
class ListComponentPipelines(Resource):
    @ns.marshal_with(component_pipelines_model)
    @ns.response(int(HTTPStatus.OK), 'Success')
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)')
    def get(self, component):
        query = SERVER.DB.session.query
        query = query(func.json_extract(SessionModel.description, '$.pipeline_branch'))
        query = query.filter(MetricModel.component == component, MetricModel.session_h == SessionModel.h)
        query = query.distinct()

        def collect(items):
            return dict(component=component, pipelines=[item[0] for item in items])

        return make_page(request, query, collector=collect)


@ns.route('/<string:component>/pipelines/<string:pipeline>/builds',
          doc=dict(description='List all builds of pipeline :pipeline which issued metrics'
                               ' for the component :component'))
class ListComponentPipelineBuilds(Resource):
    @ns.marshal_with(component_pipeline_builds_model)
    @ns.response(int(HTTPStatus.OK), 'Success')
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)')
    def get(self, component, pipeline):
        query = SERVER.DB.session.query
        query = query(func.json_extract(SessionModel.description, '$.pipeline_build_no'))
        query = query.filter(MetricModel.component == component, MetricModel.session_h == SessionModel.h,
                             func.json_extract(SessionModel.description, '$.pipeline_branch') == pipeline,
                             func.json_extract(SessionModel.description, '$.pipeline_build_no') != ''
                             ).distinct()

        def collect(items):
            return dict(component=component, pipeline=pipeline, builds=[item[0] for item in items])

        return make_page(request, query, collector=collect)


@ns.route('/<string:component>/pipelines/<string:pipeline>/builds/<string:build_no>/sessions',
          doc=dict(description='List all sessions used by build :build_no on pipeline :pipeline'
                               ' for which metrics can be found with component :component.'))
class ListSessionsOfComponentPipelineBuild(Resource):
    @ns.marshal_with(component_pipeline_build_sessions_model)
    @ns.response(int(HTTPStatus.OK), 'Success')
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)')
    def get(self, component, pipeline, build_no):
        query = SERVER.DB.session.query
        comp_session_h = MetricModel.session_h.label('comp_session_h')
        inner_query = SERVER.DB.session.query(comp_session_h)
        inner_query = inner_query.filter(MetricModel.component == component).distinct().subquery()
        query = query(inner_query.c.comp_session_h)
        query = query.filter(func.json_extract(SessionModel.description, '$.pipeline_branch') == pipeline,
                             func.json_extract(SessionModel.description, '$.pipeline_build_no') == build_no,
                             inner_query.c.comp_session_h == SessionModel.h)
        query = query.distinct()

        def collect(items):
            return dict(component=component, pipeline=pipeline, build=build_no,
                        sessions=[item[0] for item in items])

        return make_page(request, query, collector=collect)


@ns.route('/<string:component>/items/<string:item>/metrics',
          doc=dict(description="List all metrics matching item for the specified component."))
@ns.route('/items/<string:item>/metrics', defaults=dict(component=None),
          doc=dict(description="List all metrics matching item without component set."))
class ListComponentMetricsByItem(Resource):
    @ns.marshal_list_with(metrics_by_component_model, skip_none=True)
    @ns.response(int(HTTPStatus.OK), 'Success', metrics_by_component_model)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)')
    def get(self, item, component):
        if component is None:
            component = ''
        session = SERVER.DB.session
        query = session.query(MetricModel).filter(MetricModel.component == component, MetricModel.item == item)
        return make_page(request, query, 'metrics')


@ns.route('/<string:component>/variants/<string:variant>/metrics',
          doc=dict(description="List all metrics matching variant for the specified component."))
@ns.route('/variants/<string:variant>/metrics', defaults=dict(component=None),
          doc=dict(description="List all metrics matching variant without component set."))
class ListComponentMetricsByVariant(Resource):
    @ns.marshal_list_with(metrics_by_component_model, skip_none=True)
    @ns.response(int(HTTPStatus.OK), 'Success', metrics_by_component_model)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)')
    def get(self, variant, component):
        if component is None:
            component = ''
        return make_page(request,
                         MetricModel.query.filter(MetricModel.component == component,
                                                  MetricModel.item_variant == variant),
                         'metrics')



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from monitor_server import SERVER

if SERVER.config['SQLALCHEMY_DATABASE_URI'].startswith('sqlite:'):
    from sqlalchemy.dialects.sqlite.json import JSON
else:
    from sqlalchemy.dialects.postgresql.json import JSON


class MetricModel(SERVER.DB.Model):
    __tablename__ = 'TEST_METRICS'

    test_id = SERVER.DB.Column('ITEM_PK', SERVER.DB.Integer, primary_key=True)
    session_h = SERVER.DB.Column('SESSION_H', SERVER.DB.ForeignKey('TEST_SESSIONS.SESSION_H'))
    ctx_h = SERVER.DB.Column('CONTEXT_H', SERVER.DB.ForeignKey('EXECUTION_CONTEXTS.ENV_H'))
    item_start_time = SERVER.DB.Column('ITEM_START_TIME', SERVER.DB.String(64), nullable=False)
    item_path = SERVER.DB.Column('ITEM_PATH', SERVER.DB.String(4096), nullable=False)
    item = SERVER.DB.Column('ITEM', SERVER.DB.String(2048), nullable=False)
    item_variant = SERVER.DB.Column('ITEM_VARIANT', SERVER.DB.String(2048), nullable=False)
    item_fs_loc = SERVER.DB.Column('ITEM_FS_LOC', SERVER.DB.String(2048), nullable=False)
    kind = SERVER.DB.Column('KIND', SERVER.DB.String(64), nullable=False)
    component = SERVER.DB.Column('COMPONENT', SERVER.DB.String(512), nullable=True)
    wall_time = SERVER.DB.Column('TOTAL_TIME', SERVER.DB.Float, nullable=False)
    user_time = SERVER.DB.Column('USER_TIME', SERVER.DB.Float, nullable=False)
    krnl_time = SERVER.DB.Column('KERNEL_TIME', SERVER.DB.Float, nullable=False)
    cpu_usage = SERVER.DB.Column('CPU_USAGE', SERVER.DB.Float, nullable=False)
    mem_usage = SERVER.DB.Column('MEM_USAGE', SERVER.DB.Float, nullable=False)


class ExecutionContextModel(SERVER.DB.Model):
    __tablename__ = 'EXECUTION_CONTEXTS'

    h = SERVER.DB.Column('ENV_H', SERVER.DB.String(64), primary_key=True, nullable=False)
    cpu_count = SERVER.DB.Column('CPU_COUNT', SERVER.DB.Integer, nullable=False)
    cpu_freq = SERVER.DB.Column('CPU_FREQUENCY_MHZ', SERVER.DB.Integer, nullable=False)
    cpu_type = SERVER.DB.Column('CPU_TYPE', SERVER.DB.String(64), nullable=False)
    cpu_vendor = SERVER.DB.Column('CPU_VENDOR', SERVER.DB.String(256), nullable=True)
    ram_total = SERVER.DB.Column('RAM_TOTAL_MB', SERVER.DB.Integer, nullable=False)
    mac_node = SERVER.DB.Column('MACHINE_NODE', SERVER.DB.String(512), nullable=False)
    mac_type = SERVER.DB.Column('MACHINE_TYPE', SERVER.DB.String(32), nullable=False)
    mac_arch = SERVER.DB.Column('MACHINE_ARCH', SERVER.DB.String(16), nullable=False)
    sys_info = SERVER.DB.Column('SYSTEM_INFO', SERVER.DB.String(256), nullable=False)
    py_info = SERVER.DB.Column('PYTHON_INFO', SERVER.DB.String(512), nullable=False)
    ctx_h_rel = SERVER.DB.relationship('MetricModel', backref='exec_ctx', lazy=True)


class SessionModel(SERVER.DB.Model):
    __tablename__ = 'TEST_SESSIONS'

    h = SERVER.DB.Column('SESSION_H', SERVER.DB.String(64), primary_key=True, nullable=False)
    run_date = SERVER.DB.Column('RUN_DATE', SERVER.DB.String(64), nullable=False)
    scm_ref = SERVER.DB.Column('SCM_REF', SERVER.DB.String(128), nullable=True)
    description = SERVER.DB.Column('DESCRIPTION', SERVER.DB.JSON(), nullable=True)
    session_h_rel = SERVER.DB.relationship('MetricModel', backref='sessions_ctx', lazy=True)



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from flask import request
from flask_restx import Resource
from sqlalchemy import func

from http import HTTPStatus
from monitor_server import SERVER
from monitor_server.api.model import (xcontexts_ns as ns,
                                      xcontext_list_model, xcontext_model,
                                      xcontext_count_model, metrics_by_xcontext_model,
                                      metrics_by_xcontext_count_model)
from monitor_server.api.pager_utils import make_page
from monitor_server.data.model import ExecutionContextModel, MetricModel


@ns.route('/', defaults=dict(environ_hash=None),
          doc=dict(description='List all contexts with detailed information about them.'))
@ns.route('/<string:environ_hash>',
          doc=dict(description='List all contexts having their hash key starting with :environ_hash'
                               '. The listing is provided with detailed information.'))
class ListContexts(Resource):
    @ns.marshal_list_with(xcontext_list_model, skip_none=True)
    @ns.response(int(HTTPStatus.OK), 'Success', xcontext_list_model)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', xcontext_list_model)
    def get(self, environ_hash):
        if environ_hash is None:
            query = ExecutionContextModel.query
        else:
            ctx_hash = f'{environ_hash}%'
            query = ExecutionContextModel.query.filter(ExecutionContextModel.h.like(ctx_hash))
        return make_page(request, query, 'contexts')

    @ns.marshal_with(xcontext_model)
    @ns.response(int(HTTPStatus.BAD_REQUEST), 'Validation Error')
    @ns.response(int(HTTPStatus.CREATED), 'Success', xcontext_model)
    def post(self, environ_hash):
        value = ExecutionContextModel(h=request.json['h'],
                                      cpu_count=request.json['cpu_count'],
                                      cpu_freq=request.json['cpu_frequency'],
                                      cpu_type=request.json['cpu_type'],
                                      cpu_vendor=request.json['cpu_vendor'] if 'cpu_vendor' in request.json else '',
                                      ram_total=request.json['ram_total'],
                                      mac_node=request.json['machine_node'],
                                      mac_type=request.json['machine_type'],
                                      mac_arch=request.json['machine_arch'],
                                      sys_info=request.json['system_info'],
                                      py_info=request.json['python_info'])
        SERVER.DB.session.add(value)
        SERVER.DB.session.commit()
        return value, HTTPStatus.CREATED


@ns.route('/count', doc=dict(description='Count all contexts.'))
class CountContexts(Resource):
    @ns.marshal_list_with(xcontext_count_model)
    @ns.response(int(HTTPStatus.OK), 'Success', xcontext_count_model)
    def get(self):
        return dict(count=SERVER.DB.session.query(func.count(ExecutionContextModel.h)).first()[0])


@ns.route('/<string:context_hash>/metrics',
          doc=dict(description='Get all metrics linked to contexts with key :context_hash.'))
class ListMetricsByContext(Resource):
    @ns.marshal_list_with(metrics_by_xcontext_model, skip_none=True)
    @ns.response(int(HTTPStatus.OK), 'Success', metrics_by_xcontext_model)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', metrics_by_xcontext_model)
    def get(self, context_hash):
        query = MetricModel.query.filter(MetricModel.ctx_h == context_hash)
        return make_page(request, query, 'metrics')


@ns.route('/<string:context_hash>/metrics/count',
          doc=dict(description='Count the total number of metrics linked to contexts with key :context_hash.'))
class CountMetricsByContext(Resource):
    @ns.marshal_list_with(metrics_by_xcontext_count_model)
    @ns.response(int(HTTPStatus.OK), 'Success', metrics_by_xcontext_count_model)
    def get(self, context_hash):
        db_session = SERVER.DB.session
        count = db_session.query(func.count(MetricModel.test_id))
        count = count.filter(MetricModel.ctx_h == context_hash)
        return dict(count=count.first()[0]), HTTPStatus.OK



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from flask import request
from flask_restx import Resource
from sqlalchemy import func

from http import HTTPStatus
from monitor_server import SERVER
from monitor_server.api.model import (metrics_ns, items_ns, variants_ns, filters_ns,
                                      metric_model, metric_list_model, metric_count_model,
                                      item_metric_list_model, item_metrics_count_model,
                                      vart_metric_list_model, vart_metrics_count_model,
                                      filtr_metric_list_model, filtr_metrics_count_model,
                                      filtr_session_list_model)
from monitor_server.api.pager_utils import make_page
from monitor_server.data.model import MetricModel, SessionModel


@metrics_ns.route('/',
                  doc=dict(description='List all metrics with detailed information.'))
class ListMetrics(Resource):
    @metrics_ns.marshal_list_with(metric_list_model, skip_none=True)
    @metrics_ns.response(int(HTTPStatus.OK), 'Success', metric_list_model)
    @metrics_ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', metric_list_model)
    def get(self):
        return make_page(request, MetricModel.query, 'metrics')

    @metrics_ns.expect(metric_model)
    @metrics_ns.marshal_list_with(metric_model)
    @metrics_ns.response(int(HTTPStatus.CREATED), 'Success', metric_model)
    def post(self):
        metric = MetricModel(session_h=request.json['session_h'],
                             ctx_h=request.json['context_h'],
                             item_start_time=request.json['item_start_time'],
                             item_path=request.json['item_path'],
                             item=request.json['item'],
                             item_variant=request.json['item_variant'],
                             item_fs_loc=request.json['item_fs_loc'],
                             kind=request.json['kind'],
                             component=request.json['component'],
                             wall_time=request.json['total_time'],
                             user_time=request.json['user_time'],
                             krnl_time=request.json['kernel_time'],
                             cpu_usage=request.json['cpu_usage'],
                             mem_usage=request.json['mem_usage'])

        SERVER.DB.session.add(metric)
        SERVER.DB.session.commit()
        return metric, HTTPStatus.CREATED


@metrics_ns.route('/count', doc=dict(description='Count all metrics.'))
class CountMetrics(Resource):
    @metrics_ns.marshal_list_with(metric_count_model)
    @metrics_ns.response(int(HTTPStatus.OK), 'Success', metric_count_model)
    def get(self):
        db_session = SERVER.DB.session
        count = db_session.query(func.count(MetricModel.test_id))
        return dict(count=count.first()[0]), HTTPStatus.OK


@items_ns.route('/<string:item>/metrics',
                doc=dict(description='Get all metrics for test item :item. Listing contains all information'))
class ListMetricsByItem(Resource):
    @items_ns.marshal_list_with(item_metric_list_model, skip_none=True)
    @items_ns.response(int(HTTPStatus.OK), 'Success', item_metric_list_model)
    @items_ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', item_metric_list_model)
    def get(self, item):
        return make_page(request, MetricModel.query.filter(MetricModel.item == item), 'metrics')


@items_ns.route('/<string:item>/metrics/count',
                doc=dict(description='Count all metrics for the given item.'))
class CountMetricsByItem(Resource):
    @items_ns.marshal_list_with(item_metrics_count_model, skip_none=True)
    @items_ns.response(int(HTTPStatus.OK), 'Success', item_metrics_count_model)
    def get(self, item):
        db_session = SERVER.DB.session
        query = db_session.query(func.count(MetricModel.test_id))
        query = query.filter(MetricModel.item == item)
        return dict(count=query.first()[0]), HTTPStatus.OK


@items_ns.route('/like/<string:item>/metrics',
                doc=dict(description='List all metrics having test item like :item'))
class ListMetricsByItemPattern(Resource):
    @items_ns.marshal_list_with(item_metric_list_model, skip_none=True)
    @items_ns.response(int(HTTPStatus.OK), 'Success', item_metric_list_model)
    @items_ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', item_metric_list_model)
    def get(self, item):
        item = f'%{item}%'
        return make_page(request, MetricModel.query.filter(MetricModel.item.like(item)), 'metrics')


@variants_ns.route('/<string:variant>/metrics',
                   doc=dict(description='List all metric having for test variant :variant. '))
class ListMetricsByItemVariant(Resource):
    @variants_ns.marshal_list_with(vart_metric_list_model, skip_none=True)
    @variants_ns.response(int(HTTPStatus.OK), 'Success', vart_metric_list_model)
    @variants_ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', vart_metric_list_model)
    def get(self, variant):
        return make_page(request, MetricModel.query.filter(MetricModel.item_variant == variant), 'metrics')


@variants_ns.route('/like/<string:variant>/metrics',
                   doc=dict(description='List all metrics having test variant like :variant'))
class ListMetricsByItemVariantPattern(Resource):
    @variants_ns.marshal_list_with(vart_metric_list_model, skip_none=True)
    @variants_ns.response(int(HTTPStatus.OK), 'Success', vart_metric_list_model)
    @variants_ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', vart_metric_list_model)
    def get(self, variant):
        variant = f'%{variant}%'
        return make_page(request, MetricModel.query.filter(MetricModel.item_variant.like(variant)), 'metrics')


@variants_ns.route('/<string:variant>/metrics/count',
                   doc=dict(description='Count all metrics with the specified variant.'))
class CountMetricsByItemVariant(Resource):
    @variants_ns.marshal_list_with(vart_metrics_count_model, skip_none=True)
    @variants_ns.response(int(HTTPStatus.OK), 'Success', vart_metrics_count_model)
    def get(self, variant):
        db_session = SERVER.DB.session
        query = db_session.query(func.count(MetricModel.test_id))
        query = query.filter(MetricModel.item_variant == variant)
        return dict(count=query.first()[0]), HTTPStatus.OK


@filters_ns.route('/scope/<string:scope>/metrics',
                  doc=dict(description='List all metric with type set to :kind'))
class ListMetricsByType(Resource):
    @filters_ns.marshal_list_with(filtr_metric_list_model, skip_none=True)
    @filters_ns.response(int(HTTPStatus.OK), 'Success', filtr_metric_list_model)
    @filters_ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', filtr_metric_list_model)
    def get(self, scope):
        return make_page(request, MetricModel.query.filter(MetricModel.kind == scope), 'metrics')


@filters_ns.route('/scope/<string:scope>/metrics/count',
                  doc=dict(description='Count all metric with type set to :kind'))
class CountMetricsByType(Resource):
    @filters_ns.marshal_list_with(filtr_metrics_count_model, skip_none=True)
    @filters_ns.response(int(HTTPStatus.OK), 'Success', filtr_metrics_count_model)
    def get(self, scope):
        query = SERVER.DB.session.query(func.count(MetricModel.test_id))
        query = query.filter(MetricModel.kind == scope)
        return dict(count=query.first()[0]), HTTPStatus.OK


@filters_ns.route('/scm/<string:scm>/metrics/count',
                  doc=dict(description='Count all metrics related to SCM reference :scm.'))
class CountMetricsBySCMRef(Resource):
    @filters_ns.marshal_list_with(filtr_metrics_count_model)
    @filters_ns.response(int(HTTPStatus.OK), 'Success', filtr_metrics_count_model)
    def get(self, scm=None):
        db_session = SERVER.DB.session
        scm = f'{scm}%'
        sessions = [s.h for s in SessionModel.query.filter(SessionModel.scm_ref.like(scm)).all()]
        if not sessions:
            return dict(count=0)
        count = db_session.query(func.count(MetricModel.test_id)).filter(MetricModel.session_h.in_(sessions))
        return dict(count=count.first()[0]), HTTPStatus.OK


@filters_ns.route('/scm/<string:scm_id>/metrics',
                  doc=dict(description='Get all metrics with sessions having their SCM reference set to :scm_id.'))
class ListMetricsByScm(Resource):
    @filters_ns.marshal_list_with(filtr_metric_list_model, skip_none=True)
    @filters_ns.response(int(HTTPStatus.OK), 'Success', filtr_metric_list_model)
    @filters_ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', filtr_metric_list_model)
    def get(self, scm_id):
        scm = f'{scm_id}%'
        sessions = [s.h for s in SessionModel.query.filter(SessionModel.scm_ref.like(scm)).all()]
        if not sessions:
            return {'metrics': []}, HTTPStatus.NO_CONTENT
        return make_page(request, MetricModel.query.filter(MetricModel.session_h.in_(sessions)), 'metrics')


@filters_ns.route('/scm/<string:scm_id>/sessions',
                  doc=dict(description='Get all sessions with SCM reference set to :scm_id'))
class ListSessionsByScm(Resource):
    @filters_ns.marshal_list_with(filtr_session_list_model, skip_none=True)
    @filters_ns.response(int(HTTPStatus.OK), 'Success', filtr_session_list_model)
    @filters_ns.response(int(HTTPStatus.NO_CONTENT), 'Success (no content)', filtr_session_list_model)
    def get(self, scm_id):
        scm = f'{scm_id}%'
        sessions = SessionModel.query.filter(SessionModel.scm_ref.like(scm))
        return make_page(request, sessions, 'sessions')



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

import os


class Config:
    DEBUG = False
    SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory:'
    SQLALCHEMY_TRACK_MODIFICATIONS = False
    PAGINATION_COUNT = int(os.environ.get('MONITOR_SERVER_PAGINATION_COUNT', 500))
    TESTING = False


class ProdConfig(Config):
    FLASK_ENV = 'production'
    SQLALCHEMY_ECHO = bool(int(os.environ.get('MONITOR_SERVER_ECHO_SQL', 0)))
    SQLALCHEMY_DATABASE_URI = os.environ.get("MONITOR_SERVER_DATABASE_URI")


class DevConfig(Config):
    DEBUG = True
    FLASK_ENV = 'dev'
    SQLALCHEMY_ECHO = bool(int(os.environ.get('MONITOR_SERVER_ECHO_SQL', 0)))
    SERVER_NAME = os.environ.get('SERVER', '127.0.0.1:5000')
    SQLALCHEMY_DATABASE_URI = os.environ.get("MONITOR_SERVER_DATABASE_URI", 'sqlite:////tmp/pymon')


class TestConfig(Config):
    ENV = 'test'
    TESTING = False
    SQLALCHEMY_ECHO = bool(int(os.environ.get('MONITOR_SERVER_ECHO_SQL', 0)))
    PAGINATION_COUNT = int(os.environ.get('MONITOR_SERVER_PAGINATION_COUNT', 5))
    SQLALCHEMY_DATABASE_URI = 'sqlite:///:memory:'



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from monitor_api.enums import MetricScope, ResourceType, ResourceMethod
from monitor_api.core.collections import Contexts, Metrics, Sessions
from monitor_api.core.entities import Context, Metric, Session
from monitor_api.dialects.dialect import Dialect
from http import HTTPStatus
from typing import Optional, List, Dict, Any, Union
import requests


def translate(context: Optional[Dict[str, Any]] = None,
              session: Optional[Dict[str, Any]] = None,
              metric: Optional[Dict[str, Any]] = None) -> Dict[str, Any]:
    if context is not None:
        d = dict(h=context['context_h'],
                 cpu_count=context['cpu_count'],
                 cpu_freq=context['cpu_frequency'],
                 cpu_type=context['cpu_type'],
                 cpu_vendor=context['cpu_vendor'],
                 ram_total=context['ram_total'],
                 mac_node=context['machine_node'],
                 mac_type=context['machine_type'],
                 mac_arch=context['machine_arch'],
                 sys_info=context['system_info'],
                 py_info=context['python_info'])
    elif session is not None:
        d = dict(h=session['session_h'],
                 run_date=session['run_date'],
                 tags=session['tags'],
                 scm_ref=session['scm_ref']
                 )
    else:
        d = dict(context_h=metric['context_h'],
                 session_h=metric['session_h'],
                 start_time=metric['item_start_time'],
                 item_path=metric['item_path'],
                 item=metric['item'],
                 variant=metric['item_variant'],
                 path=metric['item_fs_loc'],
                 component=metric['component'],
                 wall_time=metric['total_time'],
                 user_time=metric['user_time'],
                 kernel_time=metric['kernel_time'],
                 cpu_usage=metric['cpu_usage'],
                 memory_usage=metric['mem_usage'])
        if metric['kind'].lower() == 'function':
            d.update(dict(kind=MetricScope.FUNCTION))
        elif metric['kind'].lower() == 'module':
            d.update(dict(kind=MetricScope.MODULE))
        else:
            d.update(dict(kind=MetricScope.PACKAGE))
    return d


class Remote(Dialect):
    def __init__(self, url, **kwargs):
        self.__session = requests.session()
        self.__url = url.rstrip('/')
        for i in ['headers', 'cookies', 'auth', 'proxies', 'hooks', 'params', 'verify',
                  'cert', 'adapters', 'stream', 'trust_env', 'max_redirects']:
            if i in kwargs:
                setattr(self.__session, i, kwargs[i])
        if 'adapter' in kwargs:
            self.__session.mount(self.__url, kwargs['adapter'])

    def _make_url(self, path, **params):
        url = f'{self.__url}/{path.lstrip("/")}'
        if not params:
            return url
        params = '&'.join([f'{k}={v}' for k, v in params.items()])
        return f'{url}?{params}'

    def _collect(self, url):
        r = self.__session.get(url)
        while True:
            if r.status_code == HTTPStatus.OK:
                yield r.json()
                r.status_code = HTTPStatus.NO_CONTENT
            else:
                break
            url = r.json().get('next_url', None)
            if url is not None:
                r = self.__session.get(f'{self._make_url(url)}')

    def count_sessions(self) -> int:
        try:
            j = self.__session.get(self._make_url('/sessions/count')).json()
            return j.get('count', -1)
        except requests.RequestException:
            return -1

    def count_components(self) -> int:
        try:
            j = self.__session.get(self._make_url('/components/count')).json()
            return j.get('count', -1)
        except requests.RequestException:
            return -1

    def count_contexts(self) -> int:
        try:
            j = self.__session.get(self._make_url('/contexts/count')).json()
            return j.get('count', -1)
        except requests.RequestException:
            return -1

    def count_metrics(self, session_h: str = None, context_h: str = None, scm_ref: str = None) -> int:
        try:
            if session_h:
                j = self.__session.get(self._make_url(f'/sessions/{session_h}/metrics/count')).json()
            elif context_h:
                j = self.__session.get(self._make_url(f'/contexts/{context_h}/metrics/count')).json()
            elif scm_ref:
                j = self.__session.get(self._make_url(f'/filters/scm/{scm_ref}/metrics/count')).json()
            else:
                j = self.__session.get(self._make_url('/metrics/count')).json()
            return j.get('count', -1)
        except requests.RequestException:
            return -1

    def get_session_details(self, session_h) -> Optional[Session]:
        try:
            j = self.__session.get(self._make_url(f'/sessions/{session_h}'))
            if j.status_code == HTTPStatus.NO_CONTENT:
                return None
            return Session(**translate(session=j.json()['sessions'][0]))
        except requests.RequestException:
            return None

    def get_components(self) -> List[str]:
        try:
            j = self.__session.get(self._make_url(f'/components/'))
            if j.status_code == HTTPStatus.NO_CONTENT:
                return []
            return j.json()['components']
        except requests.RequestException:
            return []

    def get_sessions(self, with_tags: Union[str, List[str]] = None,
                     restrict_flags: Union[str, List[str]] = None,
                     method: str = None) -> Sessions:
        sessions = Sessions()
        url = '/sessions/'
        params = dict()
        if with_tags:
            params['with_tags'] = ','.join(with_tags) if type(with_tags) is list else with_tags
        if restrict_flags:
            params['restrict_flags'] = ','.join(restrict_flags) if type(restrict_flags) is list else restrict_flags
        if params and method:
            params['method'] = method
        try:
            for page in self._collect(self._make_url(url, **params)):
                for session in page['sessions']:
                    s = Session(**translate(session=session))
                    sessions[s.h] = s
            return sessions
        except requests.RequestException:
            return Sessions()

    def get_metrics(self) -> Metrics:
        metrics = Metrics()
        try:
            for page in self._collect(self._make_url('/metrics/')):
                for metric in page['metrics']:
                    metrics.append(Metric(**translate(metric=metric)))
            return metrics
        except requests.RequestException:
            return Metrics()

    def get_metrics_with_context(self, context_h: str) -> Metrics:
        metrics = Metrics()
        try:
            for page in self._collect(self._make_url(f'/contexts/{context_h}/metrics')):
                for metric in page['metrics']:
                    metrics.append(Metric(**translate(metric=metric)))
            return metrics
        except requests.RequestException:
            return Metrics()

    def get_metrics_with_scm_ref(self, scm_ref: str) -> Metrics:
        metrics = Metrics()
        try:
            for page in self._collect(self._make_url(f'/filters/scm/{scm_ref}/metrics')):
                for metric in page['metrics']:
                    metrics.append(Metric(**translate(metric=metric)))
            return metrics
        except requests.RequestException:
            return Metrics()

    def get_metrics_by_type(self, metric_type: MetricScope) -> Metrics:
        metrics = Metrics()
        try:
            for page in self._collect(self._make_url(f'/filters/scope/{metric_type.value}/metrics')):
                for metric in page['metrics']:
                    metrics.append(Metric(**translate(metric=metric)))
            return metrics
        except requests.RequestException:
            return Metrics()

    def get_metrics_by_pattern(self, item: str = None, variant: str = None) -> Metrics:
        if item is not None:
            base_url = f'/items/like/{item}/metrics'
        elif variant is not None:
            base_url = f'/variants/like/{variant}/metrics'
        else:
            return Metrics()
        m = Metrics()
        try:
            for page in self._collect(self._make_url(base_url)):
                for metric in page['metrics']:
                    m.append(Metric(**translate(metric=metric)))
            return m
        except requests.RequestException:
            return Metrics()

    def get_metrics_by_session(self, session_h: str) -> Metrics:
        m = Metrics()
        try:
            for page in self._collect(self._make_url(f'/sessions/{session_h}/metrics')):
                for metric in page['metrics']:
                    m.append(Metric(**translate(metric=metric)))
            return m
        except requests.RequestException:
            return Metrics()

    def get_contexts(self) -> Contexts:
        c = Contexts()
        try:
            for page in self._collect(self._make_url('/contexts/')):
                for context in page['contexts']:
                    ctx = Context(**translate(context=context))
                    c[ctx.h] = ctx
            return c
        except requests.RequestException:
            return Contexts()

    def get_pipelines(self) -> List[str]:
        try:
            data = self.__session.get(self._make_url('/pipelines/'))
            if data.status_code == HTTPStatus.NO_CONTENT:
                return []
            return [ppl for ppl in data.json()['pipelines']]
        except requests.RequestException:
            return []

    def get_pipeline_builds(self, pipeline: str) -> List[str]:
        try:
            blds = []
            for page in self._collect(self._make_url(f'/pipelines/{pipeline}/builds')):
                blds.extend((build for build in page['builds']))
            return blds
        except requests.RequestException:
            return []

    def read_component_metrics(self, component: Optional[str] = None) -> Metrics:
        m = Metrics()
        try:
            base_url = '/components/metrics' if component is None else f'/components/{component}/metrics'
            for page in self._collect(self._make_url(base_url)):
                for metric in page['metrics']:
                    m.append(Metric(**translate(metric=metric)))
            return m
        except requests.RequestException:
            return Metrics()

    def get_component_pipelines(self, component: str) -> List[str]:
        try:
            ppls = []
            for page in self._collect(self._make_url(f'/components/{component}/pipelines')):
                ppls.extend((ppl for ppl in page['pipelines']))
            return ppls
        except requests.RequestException:
            return []

    def get_component_pipelines_build(self, component: str, pipeline: str) -> List[str]:
        try:
            builds = []
            for page in self._collect(self._make_url(f'/components/{component}/pipelines/{pipeline}/builds')):
                builds.extend((bld for bld in page['builds']))
            return builds
        except requests.RequestException:
            return []

    def get_item_metrics(self, item: str) -> Metrics:
        m = Metrics()
        try:
            for page in self._collect(self._make_url(f'/items/{item}/metrics')):
                for metric in page['metrics']:
                    m.append(Metric(**translate(metric=metric)))
            return m
        except requests.RequestException:
            return Metrics()

    def get_metrics_of_variant(self, variant: str, component: str = None) -> Metrics:
        m = Metrics()
        url = f'/variants/{variant}/metrics'
        if component:
            url = f'/components/{component}/variants/{variant}/metrics'
        try:
            for page in self._collect(self._make_url(url)):
                for metric in page['metrics']:
                    m.append(Metric(**translate(metric=metric)))
            return m
        except requests.RequestException:
            return Metrics()

    def get_context_details(self, context_h: str) -> Optional[Context]:
        try:
            data = self.__session.get(self._make_url(f'/contexts/{context_h}'))
            if data.status_code == HTTPStatus.OK:
                return Context(**translate(context=data.json()['contexts'][0]))
            return None
        except requests.RequestException:
            return None

    def get_metrics_by_resource(self, resource: ResourceType, method: ResourceMethod = ResourceMethod.TOP,
                                max_element: int = 10) -> Metrics:
        which = {res.value: res.name.lower() for res in ResourceType}
        method = "head" if method == ResourceMethod.TOP else "tail"
        url = f'/resources/{which[resource.value]}/{method}/{max_element}/metrics'
        metrics = Metrics()
        try:
            for page in self._collect(self._make_url(url)):
                for metric in page["metrics"]:
                    metrics.append(Metric(**translate(metric=metric)))
            return metrics
        except requests.RequestException:
            return Metrics()

    def get_metrics_by_component_resource(self, resource: ResourceType, component: str,
                                          method: ResourceMethod = ResourceMethod.TOP,
                                          max_element: int = 10) -> Metrics:
        which = {res.value: res.name.lower() for res in ResourceType}
        method = "head" if method == ResourceMethod.TOP else "tail"
        url = f'/resources/{which[resource.value]}/components/{component}/{method}/{max_element}/metrics'
        metrics = Metrics()
        try:
            for page in self._collect(self._make_url(url)):
                for metric in page["metrics"]:
                    metrics.append(Metric(**translate(metric=metric)))
            return metrics
        except requests.RequestException:
            return Metrics()

    def get_metrics_by_pipeline_resource(self, resource: ResourceType, pipeline: str,
                                         method: ResourceMethod = ResourceMethod.TOP,
                                         max_element: int = 10) -> Metrics:
        which = {res.value: res.name.lower() for res in ResourceType}
        method = "head" if method == ResourceMethod.TOP else "tail"
        url = f'/resources/{which[resource.value]}/pipelines/{pipeline}/{method}/{max_element}/metrics'
        metrics = Metrics()
        try:
            for page in self._collect(self._make_url(url)):
                for metric in page["metrics"]:
                    metrics.append(Metric(**translate(metric=metric)))
            return metrics
        except requests.RequestException:
            return Metrics()

    def get_metrics_by_build_resource(self, resource: ResourceType, pipeline: str, build: str,
                                      method: ResourceMethod = ResourceMethod.TOP,
                                      max_element: int = 10) -> Metrics:
        which = {res.value: res.name.lower() for res in ResourceType}
        method = "head" if method == ResourceMethod.TOP else "tail"
        url = f'/resources/{which[resource.value]}/pipelines/{pipeline}/builds/{build}/{method}/{max_element}/metrics'
        metrics = Metrics()
        try:
            for page in self._collect(self._make_url(url)):
                for metric in page["metrics"]:
                    metrics.append(Metric(**translate(metric=metric)))
            return metrics
        except requests.RequestException:
            return Metrics()

    def get_metrics_from(self, sessions: Sessions = None, contexts: Contexts = None) -> Metrics:
        m = Metrics()
        if sessions:
            for h in set(sessions.keys()):
                ms = self.get_metrics_by_session(h)
                m = Metrics.merge(m, ms)
        if contexts:
            for h in set(contexts.keys()):
                ms = self.get_metrics_with_context(h)
                m = Metrics.merge(m, ms)
        return m

    def get_sessions_from(self, metrics: Metrics) -> Sessions:
        s = Sessions()
        for metric in metrics:
            if metric.session not in s:
                s[metric.session] = self.get_session_details(metric.session)
        return s

    def get_contexts_from(self, metrics: Metrics) -> Contexts:
        c = Contexts()
        for metric in metrics:
            if metric.context not in c:
                c[metric.context] = self.get_context_details(metric.context)
        return c

    def get_sessions_from_build(self, pipeline: str, build_id: str) -> Sessions:
        url = f'/pipelines/{pipeline}/builds/{build_id}/sessions'
        sessions = Sessions()
        try:
            for page in self._collect(self._make_url(url)):
                for session in page["sessions"]:
                    sessions.add(self.get_session_details(session))
            return sessions
        except requests.RequestException:
            return Sessions()



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT


import enum


class MetricScope(enum.Enum):
    """Represent the level at which pytest-monitor collects and aggregate data.

    :FUNCTION: Metrics collected for a single test function
    :MODULE: Metrics collected at module level
    :PACKAGE: Metrics collected at package level (not working yet)
    """
    FUNCTION = 'function'
    MODULE = 'module'
    PACKAGE = 'package'


class ResourceType(enum.Enum):
    """The different types of probed resources.

    :TOTAL_TIME: Real elapsed time
    :USER_TIME: Time spent running in user mode.
    :KERNEL_TIME: Time spent running in kernel mode.
    :CPU: Amount of CPU used
    :MEMORY: Amount of memory used
    """
    TOTAL_TIME = "total_time"
    USER_TIME = "user_time"
    KERNEL_TIME = "kernel_time"
    CPU = "cpu_usage"
    MEMORY = "mem_usage"


class ResourceMethod(enum.IntEnum):
    """Represents a sorting method.

    :LOWEST: sort in ascending order
    :TOP: sort in descending order
    """
    LOWEST = 0
    TOP = 1


class Field(enum.Enum):
    """Fields that can be used to control collections export.

    :SESSION_H: Session identifier
    :CONTEXT_H: Context identifier
    :ITEM_START_TIME: UTC Time at which the item has been run
    :ITEM_PATH: Python path like to the item
    :ITEM: Test identifier
    :ITEM_VARIANT: Fully qualified test identifier
    :ITEM_FS_LOC: Path to the python file hosting the test item.
    :KIND: Test scope
    :COMPONENT: Component or project to which the test belong.
    :TOTAL_TIME: Total time spent running the test
    :USER_TIME: Time spent in user mode while running the test.
    :KERNEL_TIME: Time spent in kernel mode while running the test
    :CPU_USAGE: Cpu usage while running the test.
    :MEM_USAGE: Memory usage of the test.
    :H: Session/Context identifier.
    :SCM: Source Code Management reference
    :RUN_DATE: Session start date
    :TAGS: Session description tags
    :CPU_COUNT: Number of CPU
    :CPU_FREQUENCY_MHZ: Max frequency of CPUs
    :CPU_TYPE: Cpu types (x64, ...)
    :CPU_VENDOR: CPU Vendor
    :RAM_TOTAL_MB: Amount of RAM on the machine
    :MACHINE_NODE: Fully qualified domain name
    :MACHINE_TYPE: Machine type
    :MACHINE_ARCH: Machine architecture
    :SYSTEM_INFO: Basic information about the kernel.
    :PYTHON_INFO: Basic information about the interpreter.
    """
    SESSION_H = 'session_h'
    CONTEXT_H = 'context_h'
    ITEM_START_TIME = 'start_time'
    ITEM_PATH = 'item_path'
    ITEM = 'item'
    ITEM_VARIANT = 'variant'
    ITEM_FS_LOC = 'path'
    KIND = 'kind'
    COMPONENT = 'component'
    TOTAL_TIME = 'wall_time'
    USER_TIME = 'user_time'
    KERNEL_TIME = 'kernel_time'
    CPU_USAGE = 'cpu_usage'
    MEM_USAGE = 'memory_usage'
    H = 'h'
    SCM = 'scm'
    RUN_DATE = 'run_date'
    TAGS = 'tags'
    CPU_COUNT = 'cpu_count'
    CPU_FREQUENCY_MHZ = 'cpu_freq'
    CPU_TYPE = 'cpu_type'
    CPU_VENDOR = 'cpu_vendor'
    RAM_TOTAL_MB = 'ram'
    MACHINE_NODE = 'hostname'
    MACHINE_TYPE = 'type'
    MACHINE_ARCH = 'arch'
    SYSTEM_INFO = 'sys'
    PYTHON_INFO = 'py'


METRIC_ALL_FIELDS = [Field.CONTEXT_H, Field.SESSION_H, Field.ITEM_START_TIME,
                     Field.ITEM_PATH, Field.ITEM, Field.ITEM_FS_LOC,
                     Field.ITEM_VARIANT, Field.KIND, Field.TOTAL_TIME,
                     Field.USER_TIME, Field.KERNEL_TIME, Field.CPU_USAGE,
                     Field.MEM_USAGE, Field.COMPONENT]


SESSION_ALL_FIELDS = [Field.SESSION_H, Field.SCM, Field.RUN_DATE, Field.TAGS]


CONTEXT_ALL_FIELDS = [Field.CONTEXT_H, Field.CPU_COUNT, Field.CPU_FREQUENCY_MHZ,
                      Field.CPU_TYPE, Field.CPU_VENDOR, Field.RAM_TOTAL_MB,
                      Field.MACHINE_NODE, Field.MACHINE_TYPE, Field.MACHINE_ARCH,
                      Field.SYSTEM_INFO, Field.PYTHON_INFO]



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from flask import request
from flask_restx import Resource

from sqlalchemy import func

from http import HTTPStatus
from monitor_server import SERVER
from monitor_server.api.model import (pipelines_ns as ns, pipeline_list_model, pipeline_builds_model,
                                      pipeline_build_sessions_model)
from monitor_server.api.pager_utils import make_page
from monitor_server.data.model import SessionModel


J_PIPELINE = '$.pipeline_branch'
J_BUILD = '$.pipeline_build_no'


@ns.route('/', doc=dict(description='List all pipelines identifier.'))
class ListPipelines(Resource):
    @ns.marshal_list_with(pipeline_list_model)
    @ns.response(int(HTTPStatus.OK), 'Success', pipeline_list_model)
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (No content)', pipeline_list_model)
    def get(self):
        session = SERVER.DB.session
        pipelines = session.query(func.json_extract(SessionModel.description, J_PIPELINE)).\
            filter(func.json_extract(SessionModel.description, J_PIPELINE) != '').distinct()

        ppls = [pipeline[0] for pipeline in pipelines]
        if ppls:
            return dict(pipelines=ppls), HTTPStatus.OK
        return dict(), HTTPStatus.NO_CONTENT


@ns.route('/<string:pipeline>/builds', doc=dict(description='List all builds for pipeline :pipeline'))
class ListPipelineBuilds(Resource):
    @ns.marshal_with(pipeline_builds_model)
    @ns.response(int(HTTPStatus.OK), 'Success')
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (No content)')
    def get(self, pipeline):
        session = SERVER.DB.session
        ppl = session.query(func.json_extract(SessionModel.description, J_BUILD))
        ppl = ppl.filter(func.json_extract(SessionModel.description, J_PIPELINE) == pipeline,
                         func.json_extract(SessionModel.description, J_BUILD) != '')
        ppl = ppl.distinct()

        def collect(items):
            builds = []
            for item in items:
                builds.append(item[0])
            return dict(pipeline=pipeline, builds=builds)

        return make_page(request, ppl, None, collect)


@ns.route('/<string:pipeline>/builds/<string:build_no>/sessions',
          doc=dict(description='List all sessions for pipeline :pipeline and build :build_no'))
class ListSessionsOfBuildPipeline(Resource):
    @ns.marshal_with(pipeline_build_sessions_model)
    @ns.response(int(HTTPStatus.OK), 'Success')
    @ns.response(int(HTTPStatus.NO_CONTENT), 'Success (No content)')
    def get(self, pipeline, build_no):
        session = SERVER.DB.session
        ppl = session.query(SessionModel.h)
        ppl = ppl.filter(func.json_extract(SessionModel.description, J_PIPELINE) == pipeline,
                         func.json_extract(SessionModel.description, J_BUILD) == build_no)
        ppl = ppl.distinct()

        def collect(items):
            sessions = [item[0] for item in items]
            return dict(pipeline=pipeline, build=build_no, sessions=sessions)

        return make_page(request, ppl, None, collect)



# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT

from monitor_api.enums import MetricScope, ResourceType, ResourceMethod
from monitor_api.core.entities import Context, Session
from monitor_api.core.collections import Contexts, Metrics, Sessions
from typing import Optional, List, Union
import abc


class Dialect:  # pragma: no cover
    @abc.abstractmethod
    def get_sessions(self, with_tags: Union[str, List[str]] = None,
                     restrict_flags: Union[str, List[str]] = None,
                     method: str = None) -> Sessions:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics(self) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics_with_context(self, context_h: str) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics_with_scm_ref(self, scm_ref: str) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics_by_type(self, metric_type: MetricScope) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics_by_pattern(self, item: str = None, variant: str = None) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics_by_session(self, session_h: str) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_components(self) -> List[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_contexts(self) -> Contexts:
        raise NotImplementedError

    @abc.abstractmethod
    def count_components(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def count_contexts(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def count_metrics(self, session_h: str = None, context_h: str = None, scm_ref: str = None) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def count_sessions(self) -> int:
        raise NotImplementedError

    @abc.abstractmethod
    def get_session_details(self, session_h) -> Optional[Session]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_pipelines(self) -> List[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_pipeline_builds(self, pipeline: str) -> List[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def read_component_metrics(self, component: Optional[str] = None) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_component_pipelines(self, component: str) -> List[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_component_pipelines_build(self, component: str, pipeline: str) -> List[str]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_item_metrics(self, item: str) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_context_details(self, context_h: str) -> Optional[Context]:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics_by_resource(self, resource: ResourceType, method: ResourceMethod = ResourceMethod.TOP,
                                max_element: int = 10) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics_by_component_resource(self, resource: ResourceType,  component: str,
                                          method: ResourceMethod = ResourceMethod.TOP,
                                          max_element: int = 10) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics_by_pipeline_resource(self, resource: ResourceType,  pipeline: str,
                                         method: ResourceMethod = ResourceMethod.TOP,
                                         max_element: int = 10) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics_by_build_resource(self, resource: ResourceType, pipeline: str, build: str,
                                      method: ResourceMethod = ResourceMethod.TOP,
                                      max_element: int = 10) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics_from(self, sessions: Sessions = None, contexts: Contexts = None) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_metrics_of_variant(self, variant: str, component: str = None) -> Metrics:
        raise NotImplementedError

    @abc.abstractmethod
    def get_sessions_from_build(self, pipeline: str, build_id: str) -> Sessions:
        raise NotImplementedError



#!/usr/bin/env python

# SPDX-FileCopyrightText: 2021 Jean-Sébastien Dieu <jean-sebastien.dieu@cfm.fr>
#
# SPDX-License-Identifier: MIT
import os
import pathlib
import re
from setuptools import setup, find_packages


def read_version():
    p = pathlib.Path(__file__)
    p = p.parent / "monitor_server" / '__init__.py'
    with p.open('r') as f:
        for line in f:
            if line.startswith('__version__'):
                line = line.split('=')[1].strip()
                match = re.match(r"^['\"](\d+\.\d+\.\d+\w*)['\"]", line)
                if match:
                    return match.group(1)
    raise ValueError('Unable to compute version')


def read(fname):
    file_path = pathlib.Path(__file__).parent / fname
    with file_path.open('r', encoding='utf-8') as f:
        return f.read()


def get_requirements():
    requirements = pathlib.Path(__file__).parent / 'requirements.txt'
    if not requirements.exists() and os.environ.get('RTD_CONTEXT', 'no') == 'yes':
        requirements = pathlib.Path(__file__).parent / 'requirements-prod.txt'

    with requirements.open('r', encoding='utf-8') as f:
        packages = f.readlines()
        packages.append('wheel')
    return list(set([i for i in packages])) 


setup(
    name='monitor-server-api',
    version=read_version(),
    author='Jean-Sébastien Dieu',
    author_email='dieu.jsebastien@yahoo.com',
    maintainer='Jean-Sébastien Dieu',
    maintainer_email='dieu.jsebastien@yahoo.com',
    license='MIT',
    project_urls=dict(Source='https://github.com/CFMTech/monitor-server-api',
                      Tracker='https://github.com/CFMTech/monitor-server-api/issues'),
    url='https://monitor-server-api.readthedocs.io/',
    description='A REST API that can be used a central point for collecting metrics from pytest-monitor.',
    entry_points={'console_scripts': ['monitor-server=monitor_server.app:main']},
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    packages=find_packages('.', exclude=('tests', 'example', 'docs')),
    package_data={'monitor_server': ['templates/*']},
    include_package_data=True,
    python_requires='>=3.6',
    scripts=['scripts/launch-monitor-server.sh'],
    install_requires=get_requirements(),
    options={"bdist_wheel": {"universal": False}},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Framework :: Pytest',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6', 
        'Programming Language :: Python :: 3.7', 
        'Programming Language :: Python :: 3.8', 
        'Programming Language :: Python :: 3.9', 
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: MIT License',
    ],
)
